#!/usr/bin/env python
"""
pypozyx.structures - contains ByteStructure and derived classes that contain the Pozyx data.

The PyPozyx library is designed to work with these objects, and they abstract the low-level
approach that would otherwise be necessary in using the serial port, so these are very
recommended to use.

This package is subdivided in three groups of structures, derived from ByteStructure:
    - device: containers for device functionality. Network ID, UWB settings, device coordinates/range
    - sensor_data: containers for sensor data retrieved from the Pozyx. Magnetic, acceleration, etc.
        These also take care of the physical convertion to the respective standard units for that sensor.
    - generic: as the name suggests, generic containers. The SingleRegister object is a
        multipurpose object that can be used to read single registers of any size. Data can be
        used to create your own arbitrary packed structures.

When importing this package, you will get all of the device classes, the sensor data classes,
and both SingleRegister and Data.
"""
from pypozyx.structures.device import *
from pypozyx.structures.sensor_data import *
from pypozyx.structures.generic import SingleRegister, Data



#!/usr/bin/env python
"""pypozyx.lib - Contains core and extended Pozyx user functionality through the PozyxLib class."""

from time import sleep

from pypozyx.core import PozyxCore
from pypozyx.definitions.bitmasks import *
from pypozyx.definitions.constants import *
from pypozyx.definitions.registers import *
from pypozyx.structures.device import *
from pypozyx.structures.generic import Data, SingleRegister, dataCheck
from pypozyx.structures.sensor_data import *


class PozyxLib(PozyxCore):
    """PozyxLib

    PozyxLib
    ========

    Implements the functionality users expect from Pozyx, using the methods from PozyxCore
    to communicate and interface with Pozyx both locally and remotely.
    This does not limit itself to positioning, ranging, and reading the sensor data of
    the various Pozyx sensors, but also features an assortment of troubleshooting functions,
    abstractions of frequently used registers, UWB settings, etc.

    Unlike the Arduino library, this isn't divided into parts such as 'device functions',
    'system functions', etc, but will be in the future. For now, the Arduino library should
    work as a great reference.
    """

    def __init__(self):
        pass

    ## \addtogroup system_functions
    # @{

    def setSensorMode(self, sensor_mode, remote_id=None):
        """
        Set the Pozyx's sensor mode.

        Args:
            sensor_mode: New sensor mode. See POZYX_SENSORS_MODE register. integer sensor_mode or SingleRegister(sensor_mode).

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        if not dataCheck(sensor_mode):
            sensor_mode = SingleRegister(sensor_mode)
        assert sensor_mode[0] >= 0 and sensor_mode[
            0] <= 12, 'setSensorMode: mode %i not in range (0-12)' % sensor_mode
        status = self.setWrite(POZYX_INT_MASK, sensor_mode, remote_id)
        # legacy delay?
        sleep(POZYX_DELAY_MODE_CHANGE)
        return status

    def resetSystem(self, remote_id=None):
        """
        Resets the Pozyx device.

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        return self.useFunction(POZYX_RESET_SYS, remote_id=remote_id)

    def saveRegisters(self, registers, remote_id=None):
        """
        Saves the given registers to the Pozyx's flash memory, if these are writable registers.

        This means that upon reset, the Pozyx will use these saved values instead of the default values.
        This is especially practical when changing UWB settings of an entire network, making it unnecessary
        to re - set these when resetting or repowering a device.

        DISCLAIMER: Make sure to not abuse this function in your code, as the flash memory only has a finite
        number of writecycles available, adhere to the Arduino's mentality in using flash memory.

        Args:
            registers: Registers to save to the flash memory. Data([register1, register2, ...]) or [register1, register2, ...]
                These registers have to be writable. Saving the UWB gain is currently not working.

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        self.saveConfiguration(
            POZYX_FLASH_REGS, registers, remote_id)

    def getNumRegistersSaved(self, remote_id=None):
        """
        Obtains the number of registers saved to the Pozyx's flash memory.

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            The number of saved registers.
        """
        details = Data([0] * 20)
        if self.useFunction(POZYX_FLASH_DETAILS, data=details, remote_id=remote_id) == POZYX_FAILURE:
            return POZYX_FAILURE

        num = 0
        for i in range(20):
            for j in range(8):
                num += (details[i] >> j) & 0x1
        return num

    def isRegisterSaved(self, regAddress, remote_id=None):
        """
        Returns whether the given register is saved to the Pozyx's flash memory.

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            1 if the register is saved, 0 if it's not.
        """
        details = Data([0] * 20)
        if self.useFunction(POZYX_FLASH_DETAILS, data=details, remote_id=remote_id) == POZYX_FAILURE:
            return POZYX_FAILURE
        byte_num = regAddress / 8
        bit_num = regAddress % 8
        return (details[byte_num] >> bit_num) & 0x1

    def setConfigGPIO(self, gpio_num, mode, pull, remote_id=None):
        """
        Set the Pozyx's selected GPIO pin configuration(mode and pull).

        Args:
            gpio_num: GPIO pin number, 1 to 4.
            mode: GPIO configuration mode. integer mode or SingleRegister(mode)
            pull: GPIO configuration pull. integer pull or SingleRegister(pull)

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        if not dataCheck(mode):
            mode = SingleRegister(mode)
        if not dataCheck(pull):
            pull = SingleRegister(pull)
        assert gpio_num > 0 and gpio_num <= 4, 'setConfigGPIO: GPIO number %i not in range' % gpio_num
        assert mode[0] == POZYX_GPIO_DIGITAL_INPUT or mode[0] == POZYX_GPIO_PUSHPULL or mode[
            0] == POZYX_GPIO_OPENDRAIN, 'setConfigGPIO: wrong mode'
        assert pull[0] == POZYX_GPIO_NOPULL or mode[0] == POZYX_GPIO_PULLUP or mode[
            0] == POZYX_GPIO_PULLDOWN, 'setConfigGPIO: wrong pull/mode'

        gpio_register = POZYX_CONFIG_GPIO1 + gpio_num - 1
        mask = Data([mode[0] + (pull[0] << 3)])
        return self.setWrite(gpio_register, mask, remote_id)

    def setGPIO(self, gpio_num, value, remote_id=None):
        """
        Set the Pozyx's selected GPIO pin output.

        Args:
            gpio_num: GPIO pin number, 1 to 4
            value: GPIO output value, either HIGH(1) or LOW(0). Physically, 3.3V or 0V. integer value or SingleRegister(value).

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        if not dataCheck(value):
            value = SingleRegister(value)
        assert gpio_num >= 1 and gpio_num <= 4, 'setGPIO: GPIO number %i out of bounds' % gpio_num
        assert value[0] == 0 or value[
            0] == 1, 'setGPIO: wrong value %i' % value

        gpio_register = POZYX_GPIO1 + gpio_num - 1
        return self.setWrite(gpio_register, value, remote_id)

    def setLed(self, led_num, state, remote_id=None):
        """
        Set the Pozyx's selected LED state.

        Args:
            led_num: LED pin number, 1 to 4
            state: LED output state. Boolean. True = on and False = off, you can use POZYX_LED_ON and POZYX_LED_OFF instead.

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        assert led_num >= 1 and led_num <= 4, 'setLed: led number %i not in range 1-4' % led_num
        assert state is True or state is False, 'setLed: wrong state'

        params = Data([0x1 << (led_num - 1 + 4) |
                       ((state << led_num - 1) % 256)])
        return self.useFunction(POZYX_LED_CTRL, params, None, remote_id)

    def clearConfiguration(self, remote_id=None):
        """
        Clears the Pozyx's flash memory.

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        self.getInterruptStatus(SingleRegister())
        status = self.useFunction(
            POZYX_FLASH_RESET, remote_id=remote_id)
        if status == POZYX_FAILURE:
            print("Error clearing the flash memory")
            return status
        # give the device some time to clear the flash memory
        sleep(POZYX_DELAY_FLASH)
        return status

    def configInterruptPin(self, pin=0, mode=0, active_high=False, latch=False, remote_id=None):
        """
        Configures the interrupt pin via the POZYX_INT_CONFIG register.

        Kwargs:
            pin: The Pozyx's pin ID. 1 to 4 on anchor, 1 to 6 on tag. 0 means no pin. SingleRegister or integer.
            mode: Push-pull (0) or pull (1). SingleRegister or integer. SingleRegister or integer.
            active_high: Is the interrupt voltage active high or low. Boolean.
            latch: Is the interrupt a short pulse or latch till read? Boolean.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        if dataCheck(pin):
            pin = pin[0]
        if dataCheck(mode):
            mode = mode[0]
        assert pin <= 6 and pin >= 0, 'Error: Pin number %i is incorrect, should be between 0 and 6' % pin
        assert mode == 0 or mode == 1, 'Error: Mode is incorrect, should be 0 (PUSH-PULL) or 1 (PULL).'
        int_config = SingleRegister(
            pin + (mode << 3) + (active_high << 4) + (latch << 5))
        self.setWrite(POZYX_INT_CONFIG, int_config, remote_id)

    def getWhoAmI(self, whoami, remote_id=None):
        """
        Obtains the Pozyx's WHO_AM_I.

        Args:
            whoami: Container for the read data. SingleRegister or Data([0]).

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        return self.getRead(POZYX_WHO_AM_I, whoami, remote_id)

    def getFirmwareVersion(self, firmware, remote_id=None):
        """
        Obtains the Pozyx's firmware version.

        Args:
            firmware: Container for the read data. SingleRegister or Data([0]).

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        return self.getRead(POZYX_FIRMWARE_VER, firmware, remote_id)

    def getHardwareVersion(self, hardware, remote_id=None):
        """
        Obtains the Pozyx's hardware version.

        Args:
            hardware: Container for the read data. SingleRegister or Data([0]).

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        return self.getRead(POZYX_HARDWARE_VER, hardware, remote_id)

    def getSelftest(self, selftest, remote_id=None):
        """
        Obtains the Pozyx's selftest.

        Args:
            selftest: Container for the read data. SingleRegister or Data([0]).

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        return self.getRead(POZYX_ST_RESULT, selftest, remote_id)

    def getErrorCode(self, error_code, remote_id=None):
        """
        Obtains the Pozyx's error code.

        Args:
            error_code: Container for the read data. SingleRegister or Data([0]).

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        return self.getRead(POZYX_ERRORCODE, error_code, remote_id)

    def getInterruptStatus(self, interrupts, remote_id=None):
        """
        Obtains the Pozyx's interrupt register.

        Args:
            interrupts: Container for the read data. SingleRegister or Data([0]).

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        return self.getRead(POZYX_INT_STATUS, interrupts, remote_id)

    def getCalibrationStatus(self, calibration_status, remote_id=None):
        """
        Obtains the Pozyx's calibration status.

        Args:
            calibration_status: Container for the read data. SingleRegister or Data([0]).

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        return self.getRead(POZYX_CALIB_STATUS, calibration_status, remote_id)

    def getInterruptMask(self, mask, remote_id=None):
        """
        Obtains the Pozyx's interrupt mask.

        Args:
            mask: Container for the read data. SingleRegister or Data([0]).

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        return self.getRead(POZYX_INT_MASK, mask, remote_id)

    def getConfigModeGPIO(self, gpio_num, mode, remote_id=None):
        """
        Obtain the Pozyx's configuration mode of the selected GPIO pin.

        Args:
            gpio_num: GPIO pin number, 1 to 4.
            mode: Container for the read data. SingleRegister() or Data([0]).

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT

        See Also:
            getGPIO, getConfigPullGPIO
        """
        assert gpio_num > 0 or gpio_num <= 4, 'getConfigModeGPIO: GPIO number incorrect'
        gpio_register = POZYX_CONFIG_GPIO1 + gpio_num - 1
        status = self.getRead(gpio_register, mode, remote_id)
        mode[0] &= 0x7
        return status

    def getConfigPullGPIO(self, gpio_num, pull, remote_id=None):
        """
        Obtain the Pozyx's selected GPIO pin pull.

        Args:
            gpio_num: GPIO pin number, 1 to 4.
            pull: Container for the read data. SingleRegister() or Data([0]).

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT

        See Also:
            getGPIO, getConfigModeGPIO
        """
        assert gpio_num > 0 or gpio_num <= 4, 'getConfigPullGPIO: GPIO number incorrect'
        gpio_register = POZYX_CONFIG_GPIO1 + gpio_num - 1
        status = self.getRead(gpio_register, pull, remote_id)
        pull[0] = (pull[0] & 0x18) >> 3
        return status

    def getGPIO(self, gpio_num, value, remote_id=None):
        """
        Obtain the Pozyx's value of the selected GPIO pin, being either HIGH or LOW(physically 3.3V or 0V).

        Args:
            gpio_num: GPIO pin number, 1 to 4.
            value: Container for the read data. SingleRegister() or Data([0]).

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT

        See Also:
            getConfigPullGPIO, getConfigModeGPIO
        """
        assert gpio_num > 0 or gpio_num <= 4, 'getGPIO: GPIO number incorrect'
        gpio_register = POZYX_GPIO1 + gpio_num - 1
        return self.getRead(gpio_register, value, remote_id)

    def getSystemError(self, remote_id=None):
        """
        Returns the Pozyx's system error string.

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            string with error description

        See Also:
            getErrorCode
        """
        error_code = SingleRegister()
        self.getRead(POZYX_ERRORCODE, error_code, remote_id)

        error_codes = {POZYX_ERROR_NONE: "",
                       POZYX_ERROR_I2C_WRITE: "Error 0x01: Error writing to a register through the I2C bus",
                       POZYX_ERROR_I2C_CMDFULL: "Error 0x02: Pozyx cannot handle all the I2C commands at once",
                       POZYX_ERROR_ANCHOR_ADD: "Error 0x03: Cannot add anchor to the internal device list",
                       POZYX_ERROR_COMM_QUEUE_FULL: "Error 0x04: Communication queue is full, too many UWB messages",
                       POZYX_ERROR_I2C_READ: "Error 0x05: Error reading from a register from the I2C bus",
                       POZYX_ERROR_UWB_CONFIG: "Error 0x06: Cannot change the UWB configuration",
                       POZYX_ERROR_OPERATION_QUEUE_FULL: "Error 0x07: Pozyx cannot handle all the operations at once",
                       POZYX_ERROR_STARTUP_BUSFAULT: "Error 0x08: Internal bus error",
                       POZYX_ERROR_FLASH_INVALID: "Error 0x09: Flash memory is corrupted or invalid",
                       POZYX_ERROR_NOT_ENOUGH_ANCHORS: "Error 0x0A: Not enough anchors available for positioning",
                       POZYX_ERROR_DISCOVERY: "Error 0x0B: Error during the Discovery process",
                       POZYX_ERROR_CALIBRATION: "Error 0x0C: Error during the auto calibration process",
                       POZYX_ERROR_FUNC_PARAM: "Error 0x0D: Invalid function parameters for the register function",
                       POZYX_ERROR_ANCHOR_NOT_FOUND: "Error 0x0E: The coordinates of an anchor are not found",
                       POZYX_ERROR_FLASH: "Error 0x0F: Flash error",
                       POZYX_ERROR_MEMORY: "Error 0x10: Memory error",
                       POZYX_ERROR_RANGING: "Error 0x11: Ranging failed",
                       POZYX_ERROR_RTIMEOUT1: "Error 0x12: Ranging timeout",
                       POZYX_ERROR_RTIMEOUT2: "Error 0x13: Ranging timeout",
                       POZYX_ERROR_TXLATE: "Error 0x14: Tx was late",
                       POZYX_ERROR_UWB_BUSY: "Error 0x15: UWB is busy",
                       POZYX_ERROR_POSALG: "Error 0x16: Positioning failed",
                       POZYX_ERROR_NOACK: "Error 0x17: No Acknowledge received",
                       POZYX_ERROR_NEW_TASK: "Error 0xF1: Cannot create task",
                       POZYX_ERROR_UNRECDEV: "Error 0xFE: Hardware not recognized. Please contact support@pozyx.io",
                       POZYX_ERROR_GENERAL: "Error 0xFF: General error"}

        return error_codes.get(error_code[0], "Unknown error")

    def setInterruptMask(self, mask, remote_id=None):
        """
        Set the Pozyx's interrupt mask.

        Args:
            mask: Interrupt mask. See POZYX_INT_MASK register. integer mask or SingleRegister(mask)

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        if not dataCheck(mask):
            mask = SingleRegister(mask)
        return self.setWrite(POZYX_INT_MASK, mask, remote_id)

    def setLedConfig(self, config, remote_id=None):
        """
        Set the Pozyx's LED configuration.

        Args:
            config: LED configuration. See POZYX_CONFIG_LEDS register. integer configuration or SingleRegister(configuration)

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        if not dataCheck(config):
            config = SingleRegister(config)
        return self.setWrite(POZYX_CONFIG_LEDS, config, remote_id)

    ## @}

    ## \addtogroup positioning_functions
    # @{

    def saveAnchorIds(self, remote_id=None):
        """
        Saves the anchor IDs used in positioning to the Pozyx's flash memory.

        This means that upon reset, the Pozyx won't need to be recalibrated before performing positioning.

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        self.saveConfiguration(POZYX_FLASH_ANCHOR_IDS, remote_id=remote_id)

    def getUpdateInterval(self, ms, remote_id=None):
        """
        Obtains the Pozyx's update interval.

        Args:
            ms: Container for the read data. SingleRegister or Data([0]).

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        return self.getRead(POZYX_POS_INTERVAL, ms, remote_id)

    def getRangingProtocol(self, protocol, remote_id=None):
        """
        Obtains the Pozyx's ranging protocol

        Args:
            protocol: Container for the read protocol data. SingleRegister or Data([0])

        Kwargs:
            remote_id: Remote Pozyx ID

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        return self.getRead(POZYX_RANGE_PROTOCOL, protocol, remote_id)

    def setRangingProtocol(self, protocol, remote_id=None):
        """
        Set the Pozyx's ranging protocol.

        Args:
            protocol: the new ranging protocol. See POZYX_RANGE_PROTOCOL register. integer or SingleRegister(protocol)

        Kwargs:
            remote_id: Remote Pozyx ID

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        if not dataCheck(protocol):
            protocol = SingleRegister(protocol)
        assert protocol[0] >= 0 and protocol[0] <=2, 'setRangingProtocol: wrong protocol %i' % protocol[0]

        return self.setWrite(POZYX_RANGE_PROTOCOL, protocol, remote_id)

    def getPositionAlgorithm(self, algorithm, remote_id=None):
        """
        Obtains the Pozyx's positioning algorithm.

        Args:
            algorithm: Container for the read data. SingleRegister or Data([0]).

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        status = self.getRead(POZYX_POS_ALG, algorithm, remote_id)
        algorithm[0] &= 0xF
        return status

    def getPositionDimension(self, dimension, remote_id=None):
        """
        Obtains the Pozyx's positioning dimension.

        Args:
            dimension: Container the for read data. SingleRegister or Data([0]).

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        status = self.getRead(POZYX_POS_ALG, dimension, remote_id)
        dimension[0] = (dimension[0] & 0x30) >> 4
        return status

    def getAnchorSelectionMode(self, mode, remote_id=None):
        """
        Obtains the Pozyx's anchor selection mode.

        Args:
            mode: Container for the read data. SingleRegister or Data([0]).

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        status = self.getRead(POZYX_POS_NUM_ANCHORS, mode, remote_id)
        mode[0] = (mode[0] & 0x80) >> 7
        return status

    def getNumberOfAnchors(self, nr_anchors, remote_id=None):
        """
        Obtains the Pozyx's number of selected anchors.

        Args:
            nr_anchors: Container for the read data. SingleRegister or Data([0]).

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        status = self.getRead(POZYX_POS_NUM_ANCHORS, nr_anchors, remote_id)
        nr_anchors[0] &= 0xF
        return status

    def getOperationMode(self, mode, remote_id=None):
        """
        Obtains the Pozyx's mode of operation.

        Args:
            mode: Container for the read data. SingleRegister or Data([0]).

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        return self.getRead(POZYX_OPERATION_MODE, mode, remote_id)

    def getCoordinates(self, coordinates, remote_id=None):
        """
        Obtains the Pozyx's coordinates. These are either set manually or by positioning.

        Args:
            coordinates: Container for the read data. Coordinates().

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        return self.getRead(POZYX_POS_X, coordinates, remote_id)

    def getPositionError(self, pos_error, remote_id=None):
        """
        Obtains the Pozyx's positioning error.

        Args:
            pos_error: Container for the read data. PositionError().

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        return self.getRead(POZYX_POS_ERR_X, pos_error, remote_id)

    def getPositioningAnchorIds(self, anchors, remote_id=None):
        """
        Obtain the IDs of the anchors in the Pozyx's device list used for positioning.

        You need to make sure to know how many anchors are used, as an incorrect
        size of anchors will cause the function to fail. Use getNumberOfAnchors
        to know this number.

        Args:
            anchors: Container for the read data. DeviceList(list_size=size)
            or Data([0] * size, 'H' * size).

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT

        See Also:
            getAnchorIds, getTagIds, getDeviceIds

        Example:
            >> > list_size = SingleRegister()
            >> > self.getNumberOfAnchors(list_size)
            >> > anchor_list = DeviceList(list_size=list_size[0])
            >> > self.getPositioningAnchorIds(anchor_list)
            >> > print(anchor_list)
            '0x6720, 0x6811, 0x6891'
        """
        assert len(anchors) > 0 and len(
            anchors) <= 10, 'getPositioningAnchorIds: Anchor number out of range'
        device_list_size = SingleRegister()
        status = self.getDeviceListSize(device_list_size, remote_id)
        if len(anchors) < device_list_size[0] or status == POZYX_FAILURE:
            return POZYX_FAILURE
        return self.useFunction(POZYX_POS_GET_ANCHOR_IDS, Data([]), anchors, remote_id)

    def getDeviceRangeInfo(self, device_id, device_range, remote_id=None):
        """
        Obtain the range information of the device with selected ID in the Pozyx's device list.

        Args:
            device_id: ID of desired device whose coordinates are of interest. NetworkID()
            or Data([ID], 'H') or integer ID.
            coordinates: Container for the read data. DeviceRange().

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        if not dataCheck(device_id):
            device_id = NetworkID(device_id)
        assert device_id[0] != 0, 'getDeviceRangeInfo: device ID = 0'
        return self.useFunction(POZYX_DEVICE_GETRANGEINFO, device_id, device_range, remote_id)

    def setUpdateInterval(self, ms, remote_id=None):
        """
        Set the Pozyx's update interval in ms(milliseconds).

        Args:
            ms: Update interval in ms. integer ms or SingleRegister(ms, size=2)

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        if not dataCheck(ms):
            ms = SingleRegister(ms, size=2)
        assert ms[0] > 100 and ms[
            0] <= 600000, 'setUpdateInterval: ms not 100<ms<60000'
        return self.setWrite(POZYX_POS_INTERVAL, ms, remote_id)

    def setCoordinates(self, coordinates, remote_id=None):
        """
        Set the Pozyx's coordinates.

        Args:
            coordinates: Desired Pozyx coordinates. Coordinates() or [x, y, z].

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        if not dataCheck(coordinates):
            coordinates = Coordinates(
                coordinates[0], coordinates[1], coordinates[2])
        return self.setWrite(POZYX_POS_X, coordinates, remote_id)

    def setConfigGPIO(self, gpio_num, mode, pull, remote_id=None):
        """
        Set the Pozyx's selected GPIO pin configuration(mode and pull).

        Args:
            gpio_num: GPIO pin number, 1 to 4.
            mode: GPIO configuration mode. integer mode or SingleRegister(mode)
            pull: GPIO configuration pull. integer pull or SingleRegister(pull)

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        if not dataCheck(mode):
            mode = SingleRegister(mode)
        if not dataCheck(pull):
            pull = SingleRegister(pull)
        assert gpio_num > 0 and gpio_num <= 4, 'setConfigGPIO: GPIO number %i not in range' % gpio_num
        assert mode[0] == POZYX_GPIO_DIGITAL_INPUT or mode[0] == POZYX_GPIO_PUSHPULL or mode[
            0] == POZYX_GPIO_OPENDRAIN, 'setConfigGPIO: wrong mode'
        assert pull[0] == POZYX_GPIO_NOPULL or mode[0] == POZYX_GPIO_PULLUP or mode[
            0] == POZYX_GPIO_PULLDOWN, 'setConfigGPIO: wrong pull/mode'

        gpio_register = POZYX_CONFIG_GPIO1 + gpio_num - 1
        mask = Data([mode[0] + (pull[0] << 3)])
        return self.setWrite(gpio_register, mask, remote_id)

    def setPositionFilter(self, filter_type, filter_strength, remote_id=None):
        """
        Set the Pozyx's positioning filter.

        Note that currently only FILTER_TYPE_MOVINGAVERAGE, FILTER_TYPE_MOVINGMEDIAN and FILTER_TYPE_FIR are implemented.

        Args:
            filter_type: Positioning filter type. Integer or SingleRegister.
            filter_strength: Positioning filter strength. Integer or SingleRegister.

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        if not dataCheck(filter_strength):
            filter_strength = SingleRegister(filter_strength)
        if not dataCheck(filter_type):
            filter_type = SingleRegister(filter_type)
        assert filter_type[0] == FILTER_TYPE_MOVINGAVERAGE or filter_type[
            0] == FILTER_TYPE_MOVINGMEDIAN or filter_type[0] == FILTER_TYPE_FIR or filter_type[0] == FILTER_TYPE_NONE, 'setPositionFilter: wrong filter type'
        assert filter_strength[0] >= 0 or filter_strength[0] < 16, 'setPositionFilter: wrong strength'

        params = Data([filter_type[0] + (filter_strength[0] << 4)])
        return self.setWrite(POZYX_POS_FILTER, params, remote_id)

    def setPositionAlgorithm(self, algorithm, dimension, remote_id=None):
        """
        Set the Pozyx's positioning algorithm.

        Note that currently only POZYX_POS_ALG_UWB_ONLY and POZYX_POS_ALG_TRACKING are implemented.

        Args:
            algorithm: Positioning algorithm. integer algorithm or SingleRegister(algorithm).
            dimension: Positioning dimension. integer dimension or SingleRegister(dimension).

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        if not dataCheck(algorithm):
            algorithm = SingleRegister(algorithm)
        if not dataCheck(dimension):
            dimension = SingleRegister(dimension)
        assert algorithm[0] == POZYX_POS_ALG_UWB_ONLY or algorithm[0] == POZYX_POS_ALG_TRACKING, 'setPositionAlgorithm: wrong algorithm'
        assert dimension[0] == POZYX_3D or dimension[0] == POZYX_2D or dimension[
            0] == POZYX_2_5D, 'setPositionAlgorithm: wrong dimension'

        params = Data([algorithm[0] + (dimension[0] << 4)])
        return self.setWrite(POZYX_POS_ALG, params, remote_id)

    def setSelectionOfAnchors(self, mode, nr_anchors, remote_id=None):
        """
        Set the Pozyx's coordinates.

        Args:
            mode: Anchor selection mode. integer mode or SingleRegister(mode).
            nr_anchors: Number of anchors used in positioning. integer nr_anchors or SingleRegister(nr_anchors).

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        if not dataCheck(mode):
            mode = SingleRegister(mode)
        if not dataCheck(nr_anchors):
            nr_anchors = SingleRegister(nr_anchors)
        assert mode[0] == POZYX_ANCHOR_SEL_MANUAL or mode[
            0] == POZYX_ANCHOR_SEL_AUTO, 'setSelectionOfAnchors: wrong mode'
        assert nr_anchors[0] > 2 and nr_anchors[
            0] <= 16, 'setSelectionOfAnchors: num anchors %i not in range 3-16' % nr_anchors[0]

        params = Data([(mode[0] << 7) + nr_anchors[0]])
        return self.setWrite(POZYX_POS_NUM_ANCHORS, params, remote_id)

    def setPositioningAnchorIds(self, anchors, remote_id=None):
        """
        Set the anchors the Pozyx will use for positioning.

        Args:
            anchors: List of anchors that'll be used for positioning. DeviceList() or [anchor_id1, anchor_id2, ...]

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        if not dataCheck(anchors):
            anchors = DeviceList(anchors)
        assert len(anchors) > 0 and len(
            anchors) <= 10, 'setPositioningAnchorIds: anchor_num %i out of range' % len(anchors)

        return self.useFunction(POZYX_POS_SET_ANCHOR_IDS, anchors, None, remote_id)

    def doRanging(self, destination, device_range, remote_id=None):
        """
        Performs ranging with another destination device, resulting in range information.

        This is pretty straightforward, the range information consists of the following:
            - the timestamp of the range measurement.
            - the distance between the local / remote tag and the destination
            - the RSS, which indicates the signal strength between origin and destination.

        While in the Arduino library doRemoteRanging is used for remote ranging, this function
        follows the library's convention to add remote_id as a keyword argument. Make sure that
        the destination is on the same UWB settings as this, and to pass a DeviceRange object
        for the device_range parameter.

        For an in-action example, check the "Ready to range" tutorial on the Pozyx homepage(www.pozyx.io),
        and the ready_to_range.py example found in this library's tutorial folder.

        Args:
            destination: Network ID of the destination, to perform ranging with. integer ID or NetworkID(ID)
            device_range: Container for device range measurement data. DeviceRange object.

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        assert destination != 0, 'doRanging: destination can\'t equal zero'
        if not dataCheck(destination):
            destination = NetworkID(destination)

        int_flag = POZYX_INT_STATUS_FUNC
        if remote_id is not None:
            int_flag = POZYX_INT_STATUS_RX_DATA

        status = self.useFunction(
            POZYX_DO_RANGING, destination, Data([]), remote_id)
        if status == POZYX_SUCCESS:
            status = self.checkForFlag(int_flag, POZYX_DELAY_INTERRUPT)
            if status == POZYX_SUCCESS:
                self.getDeviceRangeInfo(destination, device_range, remote_id)
            return status
        return POZYX_FAILURE

    def doPositioning(self, position, dimension=POZYX_3D, height=Data([0], 'i'), algorithm=POZYX_POS_ALG_UWB_ONLY, remote_id=None):
        """
        Performs positioning with the Pozyx. This is probably why you're using Pozyx.

        This function only performs the positioning and doesn't take care of the previous steps
        required to get this operational, so be sure to adhere to this checklist:
            - while you can perform automatic calibration, manual calibration is much more stable and reliable.
            - when using manual calibration, add all anchors using addDevice.
            - all anchors are on the same UWB settings as the device performing positioning.
            - if you're using more than four anchors, be sure to set this with setSelectionOfAnchors.

        Basic troubleshooting:
            - try to perform ranging with all devices
            - are you using a Coordinates object for your position?
            - if you perform getDeviceListSize and subsequently getDeviceIds, are these your anchors?

        While in the Arduino library doRemotePositioning is used for remote ranging, this function
        follows the library's convention to add remote_id as a keyword argument.

        For an in-action example, check the "Ready to localize" tutorial on the Pozyx homepage(www.pozyx.io),
        and the ready_to_localize.py example found in this library's tutorial folder.

        Args:
            position: Container for the positioning coordinates. Coordinates object.

        Kwargs:
            dimension: Dimension to perform positioning in. Default 2.5D. When 2.5D, make sure height is also passed along.
            height: Height of Pozyx in 2.5D positioning. Default 0. Either integer height or Data([height], 'i').
            algorithm: Algorithm used when positioning. Default UWB only.
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        assert algorithm == POZYX_POS_ALG_UWB_ONLY or algorithm == POZYX_POS_ALG_TRACKING, 'doPositioning: wrong algorithm'
        assert dimension == POZYX_3D or dimension == POZYX_2D or dimension == POZYX_2_5D, 'doPositioning: wrong dimension'

        alg_options = Data([dimension << 4 | algorithm])
        status = self.setWrite(POZYX_POS_ALG, alg_options, remote_id)
        if dimension == POZYX_2_5D:
            if not dataCheck(height):
                height = Data([height], 'i')
            status = self.setWrite(POZYX_POS_Z, height, remote_id)

        self.getInterruptStatus(SingleRegister())

        status = self.useFunction(POZYX_DO_POSITIONING, remote_id=remote_id)
        if status != POZYX_SUCCESS:
            return POZYX_FAILURE

        if remote_id is None:
            status = self.checkForFlag(
                POZYX_INT_STATUS_POS, POZYX_DELAY_POSITIONING)
            if status == POZYX_SUCCESS:
                return self.getCoordinates(position)
            return status
        else:
            if self.waitForFlag(POZYX_INT_STATUS_RX_DATA, POZYX_DELAY_REMOTE_POSITIONING) == POZYX_SUCCESS:
                rx_info = Data([0, 0], 'HB')
                self.getRead(POZYX_RX_NETWORK_ID, rx_info)
                if rx_info[0] == remote_id and rx_info[1] == position.byte_size:
                    status = self.readRXBufferData(position)
                    # necessary to update x, y, z variables of position.
                    position.load(position.data)
                    return status
                else:
                    return POZYX_FAILURE
        return POZYX_TIMEOUT

    ## @}

    ## \addtogroup sensor_data
    # @{

    def getSensorMode(self, sensor_mode, remote_id=None):
        """
        Obtains the Pozyx's sensor mode.

        Args:
            sensor_mode: Container for the read data. SingleRegister or Data([0]).

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        return self.getRead(POZYX_SENSORS_MODE, sensor_mode, remote_id)

    def getAllSensorData(self, sensor_data, remote_id=None):
        """
        Obtains all the Pozyx's sensor data in their default units.

        Args:
            sensor_data: Container for the read data. SensorData() or RawSensorData().

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        return self.getRead(POZYX_PRESSURE, sensor_data, remote_id)

    def getPressure_Pa(self, pressure, remote_id=None):
        """
        Obtain the Pozyx's pressure sensor data in Pa(pascal).

        Args:
            pressure: Container for the read data. Data([0], 'f').

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        status = self.getRead(POZYX_PRESSURE, pressure, remote_id)
        pressure[0] = pressure[0] / POZYX_PRESS_DIV_PA
        return status

    def getAcceleration_mg(self, acceleration, remote_id=None):
        """
        Obtain the Pozyx's acceleration sensor data in mg.

        Args:
            acceleration: Container for the read data. Acceleration().

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        return self.getRead(POZYX_ACCEL_X, acceleration, remote_id)

    def getMagnetic_uT(self, magnetic, remote_id=None):
        """
        Obtain the Pozyx's magnetic sensor data in uT(microtesla).

        Args:
            magnetic: Container for the read data. Magnetic().

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        return self.getRead(POZYX_MAGN_X, magnetic, remote_id)

    def getAngularVelocity_dps(self, angular_vel, remote_id=None):
        """
        Obtain the Pozyx's angular velocity sensor data in dps(degrees per second).

        Args:
            angular_vel: Container for the read data. AngularVelocity().

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        return self.getRead(POZYX_GYRO_X, angular_vel, remote_id)

    def getEulerAngles_deg(self, euler_angles, remote_id=None):
        """
        Obtain the Pozyx's euler angles sensor data in degrees(heading, roll, pitch).

        Args:
            euler_angles: Container for the read data. EulerAngles().

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        return self.getRead(POZYX_EUL_HEADING, euler_angles, remote_id)

    def getQuaternion(self, quaternion, remote_id=None):
        """
        Obtain the Pozyx's quaternion sensor data.

        Args:
            quaternion: Container for the read data. Quaternion().

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        return self.getRead(POZYX_QUAT_W, quaternion, remote_id)

    def getLinearAcceleration_mg(self, linear_acceleration, remote_id=None):
        """
        Obtain the Pozyx's linear acceleration sensor data in mg.

        Args:
            linear_acceleration: Container for the read data. LinearAcceleration() or Acceleration().

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        return self.getRead(POZYX_LIA_X, linear_acceleration, remote_id)

    def getGravityVector_mg(self, gravity_vector, remote_id=None):
        """
        Obtain the Pozyx's gravity vector sensor data in mg.

        Args:
            gravity_vector: Container for the read data. Acceleration().

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        return self.getRead(POZYX_GRAV_X, gravity_vector, remote_id)

    def getTemperature_c(self, temperature, remote_id=None):
        """
        Obtain the Pozyx's temperature sensor data in C(celsius).

        Args:
            temperature: Container for the read data. Data([0], 'f').

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        status = self.getRead(
            POZYX_TEMPERATURE, temperature, remote_id)
        temperature[0] = temperature[0] / POZYX_TEMP_DIV_CELSIUS
        return status

    ## @}

    ## \addtogroup device_list
    # @{

    def getDeviceListSize(self, device_list_size, remote_id=None):
        """
        Obtain the size of Pozyx's list of added devices.

        Args:
            device_list_size: Container for the read data. SingleRegister() or Data([0]).

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        return self.getRead(POZYX_DEVICE_LIST_SIZE, device_list_size, remote_id)

    def getDeviceIds(self, devices, remote_id=None):
        """
        Obtain the IDs of all devices in the Pozyx's device list.

        You need to make sure to know how many devices are in the list, as an incorrect
        size of anchors will cause the function to fail. Use getDeviceListSize
        to know this number.

        Args:
            devices: Container for the read data. DeviceList(list_size=size)
            or Data([0] * size, 'H' * size).

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT

        See Also:
            getAnchorIds, getTagIds, getPositioningAnchorIds

        Example:
            >> > list_size = SingleRegister()
            >> > self.getDeviceListSize(list_size)
            >> > device_list = DeviceList(list_size=list_size[0])
            >> > self.getDeviceIds(device_list)
            >> > print(device_list)
            '0x60A0, 0x6070, 0x6891'
        """
        assert len(devices) > 0 and len(
            devices) <= 20, 'getDeviceIds: size not in range'

        list_size = SingleRegister()

        status = self.getDeviceListSize(list_size, remote_id)
        if list_size[0] < len(devices) or status == POZYX_FAILURE:
            return POZYX_FAILURE

        params = Data([0, list_size[0]])

        return self.useFunction(
            POZYX_DEVICES_GETIDS, params, devices, remote_id)

    def getAnchorIds(self, anchors, remote_id=None):
        """
        Obtain the IDs of the anchors in the Pozyx's device list.

        You need to make sure to know how many anchors are in the list, as an incorrect
        size of anchors will cause the function to fail.

        Args:
            anchors: Container for the read data. SingleRegister() or Data([0]).

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT

        See Also:
            getDeviceIds, getPositioningAnchorIds, getTagIds
        """
        assert len(anchors) > 0 and len(
            anchors) <= 20, 'getAnchorIds: size not in range'
        list_size = SingleRegister()
        status = self.getDeviceListSize(list_size, remote_id)
        if list_size[0] < len(anchors) or status == POZYX_FAILURE:
            return POZYX_FAILURE
        devices = DeviceList(list_size=list_size)
        status = self.useFunction(
            POZYX_DEVICES_GETIDS, Data([]), devices, remote_id)

        if status == POZYX_SUCCESS:
            for i in range(len(anchors)):
                anchors[i] = 0x0
            j = 0
            for i in range(list_size[0]):
                data = Data([0] * 3)
                status &= self.useFunction(
                    POZYX_DEVICE_GETINFO, NetworkID(devices[i]), data)
                # why + 1?
                if data[2] == POZYX_ANCHOR_MODE + 1:
                    anchors[j] = devices[i]
                    j += 1
            # didn't find enough anchors, so the function failed.
            if j < len(anchors):
                return POZYX_FAILURE
        return status

    def getTagIds(self, tags, remote_id=None):
        """
        Obtain the IDs of the tags in the Pozyx's device list.

        You need to make sure to know how many tags are in the list, as an incorrect
        size of tags will cause the function to fail.

        Args:
            tags: Container for the read data. SingleRegister() or Data([0]).

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT

        See Also:
            getDeviceIds, getAnchorIds, getPositioningAnchorIds
        """
        assert len(tags) > 0 and len(
            tags) <= 20, 'getTagIds: size not in range'
        list_size = SingleRegister()
        status = self.getDeviceListSize(list_size, remote_id)
        if list_size[0] < len(tags) or status == POZYX_FAILURE:
            return POZYX_FAILURE
        devices = DeviceList(list_size=list_size[0])
        status = self.useFunction(
            POZYX_DEVICES_GETIDS, Data([]), devices, remote_id)

        if status == POZYX_SUCCESS:
            for i in range(len(tags)):
                tags[i] = 0x0
            j = 0
            for i in range(list_size[0]):
                data = Data([0] * 3)
                status &= self.useFunction(
                    POZYX_DEVICE_GETINFO, NetworkID(devices[i]), data)
                # why + 1?
                if data[2] == POZYX_TAG_MODE + 1:
                    tags[j] = devices[i]
                    j += 1
            # didn't find enough tags, so the function failed.
            if j < len(tags):
                return POZYX_FAILURE
        return status

    def getDeviceCoordinates(self, device_id, coordinates, remote_id=None):
        """
        Obtain the coordinates of the device with selected ID in the Pozyx's device list.

        Args:
            device_id: ID of desired device whose coordinates are of interest. NetworkID()
            or Data([ID], 'H') or integer ID.
            coordinates: Container for the read data. Coordinates().

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        if not dataCheck(device_id):
            device_id = NetworkID(device_id)
        assert device_id[0] != 0, 'getDeviceCoordinates: device ID = 0'
        return self.useFunction(
            POZYX_DEVICE_GETCOORDS, device_id, coordinates, remote_id)

    def doDiscovery(self, discovery_type=POZYX_DISCOVERY_ANCHORS_ONLY, slots=3, slot_duration=0.01, remote_id=None):
        """
        Performs discovery on the Pozyx, which will let it discover other Pozyx devices with the same
        UWB settings in range.

        Kwargs:
            discovery_type: Type of devices to discover, defaults to discovering the anchors. POZYX_DISCOVERY_ALL_DEVICES,
                POZYX_DISCOVERY_TAGS_ONLY are alternatives.
            slots: Number of timeslots used in attempt to discover devices. Default is 3 slots.
            slot_duration: Duration in seconds of each timeslot used in the discovery process. Default is 10 ms.
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        assert discovery_type == POZYX_DISCOVERY_TAGS_ONLY or discovery_type == POZYX_DISCOVERY_ANCHORS_ONLY or discovery_type == POZYX_DISCOVERY_ALL_DEVICES, 'doDiscovery: wrong type of discovery'
        assert slots > 1 and slots < 10, 'doDiscovery: number of slots %i out of range' % slots
        assert int(slot_duration *
                   1000) > 5, 'doDiscovery: slot duration too short'

        self.getInterruptStatus(SingleRegister())
        params = Data([discovery_type, slots, int(slot_duration * 1000)])

        status = self.useFunction(
            POZYX_DEVICES_DISCOVER, params, remote_id=remote_id)
        if status == POZYX_FAILURE:
            return status
        timeout_s = slot_duration * (slots + 20)
        if remote_id is None:
            return self.checkForFlag(POZYX_INT_STATUS_FUNC, timeout_s)
        else:
            # give the remote device some time to perform its discovery.
            sleep(timeout_s)
        return status

    def doAnchorCalibration(self, dimension, num_measurements, anchors, heights=None, remote_id=None):
        """
        Performs automatic anchor calibration on the Pozyx.

        Using manual calibration over automatic calibration is highly recommended, as this will not only
        be less robust to use, the results will also be worse than a carefully accurately manually measured
        setup. Using a laser measurer for this purpose is also adviced.

        When insisting on using automatic calibration, make sure that all devices are in range and able to
        communicate with the device. Try ranging with all devices first, and make sure they're on the same
        UWB settings.

        Args:
            dimension: Dimension for the automatic calibration. When 2.5D, make sure to pass along heights as well.
            num_measurements: Number of measurements to use in calibration. The
            anchors: List of anchor IDs that will be used in the calibration. DeviceList() or [anchor_id1, anchor_id2, ...]

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        assert dimension == POZYX_2D or dimension == POZYX_2_5D, 'doAnchorCalibration: wrong dimension'
        assert num_measurements > 0, 'doAnchorCalibration: a negative number of measurements isn\'t allowed'
        assert len(anchors) >= 3 and len(
            anchors) <= 6, 'doAnchorCalibration: num anchors %i out of range 3-6' % len(anchors)
        if not dataCheck(anchors):
            anchors = DeviceList(anchors)
        if dimension == POZYX_2_5D:
            if heights is None:
                heights = [2000] * len(anchors)
            for i in range(len(anchors)):
                anchor_coordinates = Coordinates(0, 0, heights[i])
                anchor = DeviceCoordinates(anchors[i], 0x1, anchor_coordinates)
                self.addDevice(anchor, remote_id)

        self.getInterruptStatus(SingleRegister())
        params = Data([dimension, num_measurements] +
                      anchors.data, 'BB' + anchors.data_format)
        status = self.useFunction(
            POZYX_DEVICES_CALIBRATE, params, remote_id=remote_id)

        if remote_id is None:
            return self.checkForFlag(POZYX_INT_STATUS_FUNC, 25000)
        else:
            # give the remote device some time to perform calibration
            # has to be thoroughly tested
            sleep(POZYX_DELAY_CALIBRATION *
                  len(anchors) * num_measurements / 20)
        return status

    def clearDevices(self, remote_id=None):
        """
        Clears the Pozyx's device list.

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        return self.useFunction(POZYX_DEVICES_CLEAR, remote_id=remote_id)

    def addDevice(self, device_coordinates, remote_id=None):
        """
        Adds a device to the Pozyx's device list. Can be either a tag or anchor.

        Args:
            device_coordinates: Device's ID, flag, and coordinates structure. DeviceCoordinates(ID, flag, Coordinates(x, y, z)) or [ID, flag, x, y, z]

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        if not dataCheck(device_coordinates):
            device_coordinates = DeviceCoordinates(device_coordinates[0], device_coordinates[
                1], Coordinates(device_coordinates[2], device_coordinates[3], device_coordinates[4]))

        return self.useFunction(POZYX_DEVICE_ADD, device_coordinates, Data([]), remote_id)

    def saveNetwork(self, remote_id=None):
        """
        Saves the Pozyx's device list to its flash memory.

        This means that upon a reset, the Pozyx will still have the same configured device list.

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        self.saveConfiguration(POZYX_FLASH_NETWORK, remote_id=remote_id)
        self.saveRegisters([POZYX_POS_NUM_ANCHORS], remote_id)

    def configureAnchors(self, anchor_list, anchor_select=POZYX_ANCHOR_SEL_AUTO, remote_id=None):
        """
        Configures a set of anchors as the relevant anchors on a device

        Args:
            anchor_list: Python list of either DeviceCoordinates or [ID, flag, x, y, z]

        Kwargs:
            anchor_select: How to select the anchors in positioning
            remote_id: Remote Pozyx ID

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE
        """
        status = POZYX_SUCCESS
        status &= self.clearDevices(remote_id)

        for anchor in anchor_list:
            if not dataCheck(anchor):
                anchor = DeviceCoordinates(
                    anchor[0], anchor[1], Coordinates(anchor[2], anchor[3], anchor[4]))
            if anchor.flag != 0x1:
                print("ID 0x%0.4x added as tag, is this intentional?" % remote_id)
            status &= self.addDevice(anchor, remote_id)

        return status & self.setSelectionOfAnchors(anchor_select, len(anchor_list), remote_id)

    def removeDevice(self, device_id, remote_id=None):
        """
        Removes a device from the Pozyx's device list, keeping the rest of the list intact

        Args:
            device_id: ID that needs to be removed. NetworkID or integer.

        Kwargs:
            remote_id: Remote Pozyx ID

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE
        """
        if dataCheck(device_id):
            device_id = device_id[0]

        status = POZYX_SUCCESS
        list_size = SingleRegister()
        status &= self.getDeviceListSize(list_size, remote_id)
        device_list = DeviceList(list_size=list_size[0])
        status &= self.getDeviceIds(device_list, remote_id)
        if device_id not in device_list or status is POZYX_FAILURE:
            return POZYX_FAILURE
        devices = []
        for id_ in device_list:
            if id_ == device_id:
                continue
            coordinates = Coordinates()
            self.getDeviceCoordinates(id_, coordinates, remote_id)
            devices.append(DeviceCoordinates(id_, 0x1, coordinates))

        anchor_select_mode = SingleRegister()
        self.getAnchorSelectionMode(anchor_select_mode, remote_id)

        self.configureAnchors(
            devices, anchor_select=anchor_select_mode, remote_id=remote_id)

    def changeDeviceCoordinates(self, device_id, new_coordinates, remote_id=None):
        """
        Changes a device's coordinates in the Pozyx's device list, keeping the rest of the list intact

        Args:
            device_id: ID that needs to be removed. NetworkID or integer.
            new_coordinates: new coordinates for the device

        Kwargs:
            remote_id: Remote Pozyx ID

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE
        """
        if dataCheck(device_id):
            device_id = device_id[0]
        if not dataCheck(new_coordinates):
            new_coordinates = Coordinates(x=new_coordinates[0], y=new_coordinates[
                                          1], z=new_coordinates[2])
        status = POZYX_SUCCESS
        list_size = SingleRegister()
        status &= self.getDeviceListSize(list_size, remote_id)
        device_list = DeviceList(list_size=list_size[0])
        status &= self.getDeviceIds(device_list, remote_id)
        if device_id not in device_list or status is not POZYX_SUCCESS:
            return POZYX_FAILURE
        devices = []
        for id_ in device_list:
            if id_ == device_id:
                devices.append(DeviceCoordinates(id_, 0x1, new_coordinates))
                continue
            coordinates = Coordinates()
            self.getDeviceCoordinates(id_, coordinates, remote_id)
            devices.append(DeviceCoordinates(id_, 0x1, coordinates))

        anchor_select_mode = SingleRegister()
        self.getAnchorSelectionMode(anchor_select_mode, remote_id)

        self.configureAnchors(
            devices, anchor_select=anchor_select_mode, remote_id=remote_id)

    def printDeviceList(self, remote_id=None):
        """
        Prints a Pozyx's device list.

        Kwargs:
            remote_id: Remote Pozyx ID

        Returns:
            None
        """
        list_size = SingleRegister()
        status = self.getDeviceListSize(list_size, remote_id)
        device_list = DeviceList(list_size=list_size[0])
        status &= self.getDeviceIds(device_list, remote_id)

        if status is POZYX_FAILURE:
            print("Couldn't read device list of Pozyx")
            return

        for device_id in device_list:
            coordinates = Coordinates()
            self.getDeviceCoordinates(device_id, coordinates, remote_id)
            print(DeviceCoordinates(device_id, 0x1, coordinates))

    ## @}

    ## \addtogroup communication_functions
    # @{

    def saveUWBSettings(self, remote_id=None):
        """
        Saves the Pozyx's UWB settings to its flash memory.

        This means that upon a reset, the Pozyx will still have the same configured UWB settings.
        As of writing, POZYX_UWB_GAIN is not savable yet.

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        registers = [POZYX_UWB_CHANNEL, POZYX_UWB_RATES,
                     POZYX_UWB_PLEN, POZYX_UWB_GAIN]
        self.saveRegisters(registers, remote_id)


    def setNetworkId(self, network_id, remote_id=None):
        """
        Set the Pozyx's network ID.

        If using this remotely, make sure to change the network ID to the new ID in
        subsequent code, as its ID will have changed and using the old ID will not work.

        Args:
            network_id: New Network ID. integer ID or NetworkID(ID) or SingleRegister(ID, size=2)

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        if not dataCheck(network_id):
            network_id = NetworkID(network_id)
        return self.setWrite(POZYX_NETWORK_ID, network_id, remote_id)

    def setUWBSettings(self, UWB_settings, remote_id=None):
        """
        Set the Pozyx's UWB settings.

        If using this remotely, remember to change the local UWB settings as well
        to make sure you are still able to communicate with the remote device.

        Args:
            UWB_settings: The new UWB settings. UWBSettings() or [channel, bitrate, prf, plen, gain_db]

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        if not dataCheck(UWB_settings):
            UWB_settings = UWBSettings(UWB_settings[0], UWB_settings[1],
                                       UWB_settings[2], UWB_settings[3], UWB_settings[4])
        gain = Data([UWB_settings.gain_db], 'f')
        UWB = Data([UWB_settings.channel, UWB_settings.bitrate +
                    (UWB_settings.prf << 6), UWB_settings.plen])
        status = self.setWrite(POZYX_UWB_CHANNEL, UWB, remote_id,
                               2 * POZYX_DELAY_LOCAL_WRITE, 2 * POZYX_DELAY_REMOTE_WRITE)
        if status == POZYX_FAILURE:
            return status
        return self.setTxPower(gain, remote_id)

    def setUWBChannel(self, channel_num, remote_id=None):
        """
        Set the Pozyx's UWB channel.

        If using this remotely, remember to change the local UWB channel as well
        to make sure you are still able to communicate with the remote device.

        Args:
            channel_num: The new UWB channel, being either 1, 2, 3, 4, 5 or 7.
                See POZYX_UWB_CHANNEL register. integer channel or SingleRegister(channel)

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        if not dataCheck(channel_num):
            channel_num = SingleRegister(channel_num)
        assert channel_num[0] >= 1 and channel_num[0] <= 7 and channel_num[
            0] != 6, 'setUWBChannel: %i is wrong channel number' % channel_num[0]

        return self.setWrite(POZYX_UWB_CHANNEL, channel_num, remote_id)

    def setTxPower(self, txgain_dB, remote_id=None):
        """
        Set the Pozyx's UWB transceiver gain.

        Args:
            txgain_dB: The new transceiver gain in dB, a value between 0.0 and 33.0.
                float gain or Data([gain], 'f').

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        if not dataCheck(txgain_dB):
            txgain_dB = Data([txgain_dB], 'f')
        assert txgain_dB[0] >= 0.0 and txgain_dB[
            0] <= 35.0, 'setTxPower: TX gain %0.2fdB not in range (0-35dB)' % txgain_dB[0]
        doublegain_dB = Data([int(2.0 * txgain_dB[0] + 0.5)])

        return self.setWrite(POZYX_UWB_GAIN, doublegain_dB, remote_id)

    def getNetworkId(self, network_id):
        """
        Obtains the Pozyx's network ID.

        Args:
            network_id: Container for the read data.  NetworkID() or SingleRegister(size=2) or Data([0], 'H').

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE
        """
        return self.regRead(POZYX_NETWORK_ID, network_id)

    def getUWBSettings(self, UWB_settings, remote_id=None):
        """
        Obtains the Pozyx's UWB settings.

        Args:
            UWB_settings: Container for the read data.  UWBSettings().

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        # The UWB data register size is 4.
        tmp_data = Data([0] * 4)
        status = self.getRead(POZYX_UWB_CHANNEL, tmp_data, remote_id)
        UWB_settings.load(tmp_data.data)
        return status

    def getUWBChannel(self, channel_num, remote_id=None):
        """
        Obtains the Pozyx's UWB channel.

        Args:
            channel_num: Container for the read data. SingleRegister or Data([0]).

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        channel_num[0] = 0
        status = self.getRead(POZYX_UWB_CHANNEL, channel_num, remote_id)
        if channel_num[0] == 0 or status == POZYX_FAILURE:
            return POZYX_FAILURE
        return status

    def getTxPower(self, txgain_dB, remote_id=None):
        """
        Obtains the Pozyx's transmitter UWB gain in dB, as a float.

        Args:
            txgain_dB: Container for the read data. Data([0], 'f').

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        doublegain_dB = SingleRegister()
        status = self.getRead(
            POZYX_UWB_GAIN, doublegain_dB, remote_id)
        txgain_dB[0] = 0.5 * doublegain_dB[0]
        return status

    def getLastNetworkId(self, network_id, remote_id=None):
        """
        Obtain the network ID of the last device Pozyx communicated with.

        Args:
            network_id: Container for the read data. NetworkID() or SingleRegister(size=2) or Data([0], 'H').

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        return self.getRead(POZYX_RX_NETWORK_ID, network_id, remote_id)

    def getLastDataLength(self, data_length, remote_id=None):
        """
        Obtain the size of the most recent data packet received by the Pozyx.

        Args:
            data_length: Container for the read data. SingleRegister() or Data([0]).

        Kwargs:
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        return self.getRead(POZYX_RX_DATA_LEN, data_length, remote_id)

    ## @}

    def saveConfiguration(self, save_type, registers=None, remote_id=None):
        """
        General function to save the Pozyx's configuration to its flash memory.

        This constitutes three different Pozyx configurations to save, and each have their specialised derived function:
            POZYX_FLASH_REGS: This saves the passed Pozyx registers if they're writable, see saveRegisters.
            POZYX_FLASH_ANCHOR_IDS: This saves the anchors used during positioning, see saveAnchorIds.
            POZYX_FLASH_NETWORK: This saves the device list to the Pozyx device, see saveNetwork.

        It is recommended to use the derived functions, as these are not just easier to use, but also
        more descriptive than this general save function.

        DISCLAIMER: Make sure to not abuse this function in your code, as the flash memory only has a finite
        number of writecycles available, adhere to the Arduino's mentality in using flash memory.

        Args:
            save_type: Type of configuration to save. See above.

        Kwargs:
            registers: Registers to save to the flash memory. Data([register1, register2, ...]) or [register1, register2, ...]
                These registers have to be writable. Saving the UWB gain is currently not working.
            remote_id: Remote Pozyx ID.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        if registers is None:
            registers = Data([])
        if not dataCheck(registers):
            registers = Data(registers)
        assert save_type == POZYX_FLASH_REGS or save_type == POZYX_FLASH_ANCHOR_IDS or save_type == POZYX_FLASH_NETWORK, 'saveConfiguration: invalid type'
        assert save_type == POZYX_FLASH_REGS or len(
            registers) == 0, 'saveConfiguration: #regs > 0 and not a reg save'
        assert save_type != POZYX_FLASH_REGS or len(
            registers) > 0, 'saveConfiguration: #regs > 0 and not a reg save'

        self.getInterruptStatus(SingleRegister())
        params = Data([save_type] + registers.data)
        status = self.useFunction(
            POZYX_FLASH_SAVE, params, remote_id=remote_id)
        if status == POZYX_FAILURE:
            print("Error saving to flash memory")
            return status
        # give the device some time to save to flash memory
        sleep(POZYX_DELAY_FLASH)
        return status



#!/usr/bin/env python
""" pypozyx.structures.byte_structure - contains the ByteStructure class, thank you struct."""

import struct


class ByteStructure(object):
    """
    The ByteStructure class is the base class that all custom structs inherit
    their basic functionality from. It implements the low-level functionality
    that makes it easy to make use arbitrary struct formats in the interface
    with Pozyx.

    TODO: Refactor then document
    """
    byte_size = 4
    data_format = 'BBBB'

    def __init__(self):
        self.byte_data = '00' * self.byte_size
        self.bytes_to_data()

    def load_bytes(self, byte_data):
        self.byte_data = byte_data
        self.bytes_to_data()

    def bytes_to_data(self):
        s = ''.encode()
        for i in range(int(len(self.byte_data) / 2)):
            index = 2 * i
            s += struct.pack('B',
                             int(self.byte_data[index:index + 2], 16))
        self.load_packed(s)

    def load_packed(self, s):
        index = 0
        self.data = [0] * len(self.data_format)
        for i in range(len(self.data_format)):
            data_len = struct.calcsize(self.data_format[i])
            self.data[i] = struct.unpack(self.data_format[i], s[
                                         index:index + data_len])[0]
            index += data_len
        self.load(self.data)

    def load_hex_string(self):
        byte_data = self.transform_to_bytes()
        s = ''
        for i in range(len(byte_data)):
            s += '%0.2x' % byte_data[i]
        self.byte_data = s

    def transform_to_bytes(self):
        new_format = ''
        for i in range(len(self.data)):
            new_format += 'B' * struct.calcsize(self.data_format[i])
        return self.transform_data(new_format)

    def set_packed_size(self):
        new_format = ''
        for i in range(len(self.data)):
            new_format += 'B' * struct.calcsize(self.data_format[i])
        self.byte_size = 1 * len(new_format)

    def set_unpacked_size(self):
        self.byte_size = struct.calcsize(self.data_format)

    def transform_data(self, new_format):
        s = ''.encode()
        for i in range(len(self.data)):
            s += struct.pack(self.data_format[i], self.data[i])
        return list(struct.unpack(new_format, s))

    def change_data(self, index, new_data):
        if type(new_data) == int:
            self.data[index] = new_data
        elif type(new_data) == list:
            for i in range(len(new_data)):
                self.data[index + i] == new_data[i]
        else:
            print("Trying to change data with invalid new values (use int or list)")

    def load(self, data, convert=1):
        """Loads data in its relevant class components."""
        raise NotImplementedError(
            'load(data) should be customised for every derived structure')

    # what is this doing here?
    def update_data(self):
        """Updates the class's data when one of its components has changed."""
        pass

    # Python Data Object functionality.
    def __getitem__(self, key):
        return self.data[key]

    def __setitem__(self, key, value):
        self.data[key] = value

    def __len__(self):
        return len(self.data)

    def __str__(self):
        """Returns a string that should be tailored to each ByteStructure-derived class."""
        s = ''
        for i in range(len(self.data)):
            if i > 0:
                s += ', '
            s += str(self.data[i])
        return s

    # should always be tailored to specific data structure.
    def __setattr__(self, name, value):
        # keep the regular assignment in place
        self.__dict__[name] = value
        # keep the data up to date when setting a variable.
        self.update_data()



#!/usr/bin/env python
"""
pypozyx.definitions - contains all Pozyx constant definitions.

These definitions are divided in three large groups:
    - constants: physical constants, type definitions...
    - bitmasks: bitmasks used for ANDing register data against, especially useful for interrupt status data.
    - registers: definitions of the register addresses, for more 'advanced' use.

When importing this, only the constants will get imported. To import the registers and bitmasks,
use the following code:
    >>> from pypozyx.definitions.registers import * # or specific registers for best practice
    >>> from pypozyx.definitions.bitmasks import * # or specific bitmasks for best practice
"""
from pypozyx.definitions.constants import *



#!/usr/bin/env python
"""
pypozyx.structures.device - contains various classes representing device data

Structures contained
--------------------
DeviceCoordinates
    consists of a device's ID, flag, and coordinates
DeviceRange
    consists of a range measurements timestamp, distance, and RSS
NetworkID
    container for a device's ID. Prints in 0xID format.
DeviceList
    container for a list of IDs. Can be initialized through size and/or IDs.
UWBSettings
    contains all of the UWB settings: channel, bitrate, prf, plen, and gain.
"""

from pypozyx.definitions.constants import *
from pypozyx.structures.byte_structure import ByteStructure
from pypozyx.structures.generic import Data
from pypozyx.structures.sensor_data import Coordinates


class DeviceCoordinates(ByteStructure):
    """
    Container for both reading and writing device coordinates from and to Pozyx.

    The keyword arguments are at once its properties.

    Kwargs:
        network_id: Network ID of the device
        flag: Type of the device. Tag or anchor.
        pos: Coordinates of the device. Coordinates().
    """
    byte_size = 15
    data_format = 'HBiii'

    def __init__(self, network_id=0, flag=0, pos=Coordinates()):
        """
        Initializes the DeviceCoordinates object.

        Kwargs:
            network_id: Network ID of the device
            flag: Type of the device. Tag or anchor.
            pos: Coordinates of the device. Coordinates().
        """
        self.network_id = network_id
        self.flag = flag
        self.pos = pos
        self.data = [network_id, flag, int(pos.x), int(pos.y), int(pos.z)]

    def load(self, data):
        self.data = data
        self.network_id = data[0]
        self.flag = data[1]
        self.pos = Coordinates(data[2], data[3], data[4])

    def update_data(self):
        try:
            if self.data != [self.network_id, self.flag,
                             self.pos.x, self.pos.y, self.pos.z]:
                self.data = [self.network_id, self.flag,
                             self.pos.x, self.pos.y, self.pos.z]
        except:
            return

    def __str__(self):
        return "ID: 0x{self.network_id:x}, flag: {self.flag}, ".format(self=self) + str(self.pos)


class DeviceRange(ByteStructure):
    """
    Container for the device range data, resulting from a range measurement.

    The keyword arguments are at once its properties.

    Kwargs:
        timestamp: Timestamp of the range measurement
        distance: Distance measured by the device.
        RSS: Signal strength during the ranging measurement.
    """
    byte_size = 10
    data_format = 'IIh'

    def __init__(self, timestamp=0, distance=0, RSS=0):
        """Initializes the DeviceRange object."""
        self.timestamp = timestamp
        self.distance = distance
        self.RSS = RSS
        self.data = [timestamp, distance, RSS]

    def load(self, data):
        self.data = data
        self.timestamp = data[0]
        self.distance = data[1]
        self.RSS = data[2]

    def update_data(self):
        try:
            if self.data != [self.timestamp, self.distance, self.RSS]:
                self.data = [self.timestamp, self.distance, self.RSS]
        except:
            return

    def __str__(self):
        return '{self.timestamp}ms, {self.distance}mm, {self.RSS}dBm'.format(self=self)


class NetworkID(Data):
    """
    Container for a device's network ID.

    Kwargs:
        network_id: The network ID of the device.
    """

    def __init__(self, network_id=0):
        """Initializes the NetworkID object."""
        Data.__init__(self, [network_id], 'H')
        self.id = network_id

    def load(self, data):
        self.data = data
        self.id = data[0]

    def update_data(self):
        try:
            if self.data != [self.id]:
                self.data = [self.id]
        except:
            return

    def __str__(self):
        return "0x%0.4x" % self.id


class DeviceList(Data):
    """
    Container for a list of device IDs.

    Using list_size is recommended when having just used getDeviceListSize, while ids
    is recommended when one knows the IDs. When using one, the other automatically
    gets its respective value. Therefore, only use on of both.

    Note also that DeviceList(list_size=1) is the same as NetworkID().

    Kwargs:
        ids: Array of known or unknown device IDs. Empty by default.
        list_size: Size of the device list.
    """

    def __init__(self, ids=[], list_size=0):
        """Initializes the DeviceList object with either IDs or its size."""
        if list_size != 0 and ids == []:
            Data.__init__(self, [0] * list_size, 'H' * list_size)
        else:
            Data.__init__(self, ids, 'H' * len(ids))

    def __str__(self):
        s = 'IDs: '
        for i in range(len(self)):
            if i > 0:
                s += ', '
            s += '0x%0.4x' % self[i]
        return s

    def load(self, data):
        for i in range(len(data)):
            self.data[i] = data[i]


class UWBSettings(ByteStructure):
    """
    Container for a device's UWB settings.

    Its keyword arguments are at once its properties.

    It also provides parsing functions for all its respective properties,
    which means this doesn't need to be done by users. These functions are
    parse_prf, parse_plen and parse_bitrate.

    You can also directly print the UWB settings, resulting in the following
    example output:
    "CH: 1, bitrate: 850kbit/s, prf: 16MHz, plen: 1024 symbols, gain: 15.0dB"

    Kwargs:
        channel: UWB channel of the device. See POZYX_UWB_CHANNEL.
        bitrate: Bitrate of the UWB commmunication. See POZYX_UWB_RATES.
        prf: Pulse repeat frequency of the UWB. See POZYX_UWB_RATES.
        plen: Preamble length of the UWB packets. See POZYX_UWB_PLEN.
        gain_db: Gain of the UWB transceiver, a float value. See POZYX_UWB_GAIN.
    """
    byte_size = 7
    data_format = 'BBBBf'

    def __init__(self, channel=0, bitrate=0, prf=0, plen=0, gain_db=0.0):
        """Initializes the UWB settings."""
        self.channel = channel
        self.bitrate = bitrate
        self.prf = prf
        self.plen = plen
        self.gain_db = float(gain_db)
        self.data = [self.channel, self.bitrate,
                     self.prf, self.plen, self.gain_db]

    def load(self, data):
        self.channel = data[0]
        self.bitrate = data[1] & 0x3F
        self.prf = (data[1] & 0xC0) >> 6
        self.plen = data[2]
        self.gain_db = float(data[3]) / 2
        self.data = [self.channel, self.bitrate,
                     self.prf, self.plen, self.gain_db]

    def update_data(self):
        try:
            if self.data != [self.channel, self.bitrate,
                             self.prf, self.plen, self.gain_db]:
                self.data = [self.channel, self.bitrate,
                             self.prf, self.plen, self.gain_db]
        except:
            return

    def parse_bitrate(self):
        """Parses the bitrate to be humanly readable."""
        bitrates = {0: '110kbit/s', 1: '850kbit/s', 2: '6.81Mbit/s'}
        try:
            return bitrates[self.bitrate]
        except:
            return 'invalid bitrate'

    def parse_prf(self):
        """Parses the pulse repetition frequency to be humanly readable."""
        prfs = {1: '16 MHz', 2: '64 MHz'}
        try:
            return prfs[self.prf]
        except:
            return 'invalid pulse repetitions frequency (PRF)'

    def parse_plen(self):
        """Parses the preamble length to be humanly readable."""
        plens = {0x0C: '4096 symbols', 0x28: '2048 symbols', 0x18: '1536 symbols', 0x08: '1024 symbols',
                 0x34: '512 symbols', 0x24: '256 symbols', 0x14: '128 symbols', 0x04: '64 symbols'}
        try:
            return plens[self.plen]
        except:
            return 'invalid preamble length'

    def __str__(self):
        return "CH: {}, bitrate: {}, prf: {}, plen: {}, gain: {}dB".format(self.channel, self.parse_bitrate(), self.parse_prf(), self.parse_plen(), self.gain_db)



#!/usr/bin/env python
"""

"""

from pypozyx.lib import PozyxLib
from pypozyx.definitions.constants import *
from pypozyx.definitions.registers import *


class PozyxI2C(PozyxLib):

    def __init__(self, mode=MODE_POLLING, print_output=False):
        pass

    def regWrite(self, address, data):
        pass

    def regRead(self, address, data):
        pass

    def regFunction(self, address, params, data):
        pass

    #
    def waitForFlag(self, interrupt_flag, timeout_ms, interrupt):
        pass

    #
    def waitForFlag_safe(self, interrupt_flag, timeout_ms, interrupt):
        pass

    def configInterruptPin(self, pin, mode, bActiveHigh, bLatch, remote_id=None):
        pass



#!/usr/bin/env python
"""pypozyx.structures.sensor_data - Contains container classes for data from the Pozyx's many sensors."""

from pypozyx.definitions.constants import *
from pypozyx.structures.byte_structure import ByteStructure
from pypozyx.structures.generic import XYZ


class Coordinates(XYZ):
    """Container for coordinates in x, y, and z (in mm)."""
    byte_size = 12
    data_format = 'iii'

    def load(self, data, convert=0):
        self.data = data
        self.x = data[0]
        self.y = data[1]
        self.z = data[2]


class Acceleration(XYZ):
    """Container for acceleration in x, y, and z (in mg)."""
    physical_convert = POZYX_ACCEL_DIV_MG

    byte_size = 6
    data_format = 'hhh'


class Magnetic(XYZ):
    """Container for coordinates in x, y, and z (in uT)."""
    physical_convert = POZYX_MAG_DIV_UT

    byte_size = 6
    data_format = 'hhh'


class AngularVelocity(XYZ):
    """Container for angular velocity in x, y, and z (in dps)."""
    physical_convert = POZYX_GYRO_DIV_DPS

    byte_size = 6
    data_format = 'hhh'


class LinearAcceleration(XYZ):
    """Container for linear acceleration in x, y, and z (in mg), as floats."""
    physical_convert = POZYX_ACCEL_DIV_MG

    byte_size = 6
    data_format = 'hhh'

    def load(self, data, convert=1):
        if convert:
            self.x = float(data[0]) / self.physical_convert
            self.y = float(data[1]) / self.physical_convert
            self.z = float(data[2]) / self.physical_convert
        else:
            self.x = float(data[0])
            self.y = float(data[1])
            self.z = float(data[2])


class PositionError(XYZ):
    """Container for position error in x, y, z, xy, xz, and yz (in mm)."""
    physical_convert = 1
    byte_size = 12
    data_format = 'hhhhhh'

    def __init__(self, x=0, y=0, z=0, xy=0, xz=0, yz=0):
        """Initializes the PositionError object."""
        XYZ.__init__(self, x, y, z)
        self.xy = xy
        self.xz = xz
        self.yz = yz
        self.data = [x, y, z, xy, xz, yz]

    def load(self, data):
        XYZ.load(self, data[0:3])
        self.xy = data[3] / self.physical_convert
        self.xz = data[4] / self.physical_convert
        self.yz = data[5] / self.physical_convert

    def update_data(self):
        try:
            if self.data != [self.x, self.y, self.z, self.xy, self.xz, self.yz]:
                self.data = [self.x, self.y, self.z, self.xy, self.xz, self.yz]
        except:
            return

    def __str__(self):
        return 'X: {self.x}, Y: {self.y}, Z: {self.z}, XY: {self.xy}, XZ: {self.xz}, YZ: {self.yz}'.format(self=self)


class Quaternion(XYZ):
    """Container for quaternion data in x, y, z and w."""
    physical_convert = POZYX_QUAT_DIV

    byte_size = 8
    data_format = 'hhhh'

    def __init__(self, w=0, x=0, y=0, z=0):
        """Initializes the Quaternion object."""
        XYZ.__init__(self, x, y, z)
        self.data = [w, x, y, z]
        self.w = w

    def load(self, data, convert=1):
        for i in range(len(data)):
            data[i] = float(data[i])
        XYZ.load(self, data[1:4], convert)
        self.data = data
        if convert:
            self.w = data[0] / self.physical_convert
        else:
            self.w = data[0]

    def update_data(self):
        try:
            if self.data != [self.w, self.x, self.y, self.z]:
                self.data = [self.w, self.x, self.y, self.z]
        except:
            return

    def __str__(self):
        return 'X: {self.x}, Y: {self.y}, Z: {self.z}, W: {self.w}'.format(self=self)


class EulerAngles(ByteStructure):
    """Container for euler angles as heading, roll, and pitch (in degrees)."""
    physical_convert = POZYX_EULER_DIV_DEG

    byte_size = 6
    data_format = 'hhh'

    def __init__(self, heading=0, roll=0, pitch=0):
        """Initializes the EulerAngles object."""
        self.data = [heading, roll, pitch]
        self.heading = heading
        self.roll = roll
        self.pitch = pitch

    def load(self, data, convert=1):
        self.data = data
        if convert:
            self.heading = float(data[0]) / self.physical_convert
            self.roll = float(data[1]) / self.physical_convert
            self.pitch = float(data[2]) / self.physical_convert
        else:
            self.heading = float(data[0])
            self.roll = float(data[1])
            self.pitch = float(data[2])

    def update_data(self):
        try:
            if self.data != [self.heading, self.roll, self.pitch]:
                self.data = [self.heading, self.roll, self.pitch]
        except:
            return

    def __str__(self):

        return 'Heading: {self.heading}, Roll: {self.roll}, Pitch: {self.pitch}'.format(
            self=self)


class SensorData(ByteStructure):
    """
    Container for all sensor data.

    This includes, in order, with respective structure:
        - pressure : float
        - acceleration : Acceleration
        - magnetic : Magnetic
        - angular_vel : AngularVelocity
        - euler_angles : EulerAngles
        - quaternion : Quaternion
        - linear_acceleration: LinearAcceleration
        - gravity_vector: LinearAcceleration
        - temperature: Int8
    """
    byte_size = 49  # 4 + 6 + 6 + 6 + 6 + 8 + 6 + 6 + 1

    # 'I' + 'h'*3 + 'h'*3 + 'h'*3 + 'h'*3 + 'f'*4 + 'h'*3 + 'h'*3 + 'B'
    data_format = 'fhhhhhhhhhhhhhhhhhhhhhhB'

    def __init__(self, data=[0] * 24):
        """Initializes the SensorData object."""
        self.data = data
        self.pressure = data[0]
        self.acceleration = Acceleration(data[1], data[2], data[3])
        self.magnetic = Magnetic(data[4], data[5], data[6])
        self.angular_vel = AngularVelocity(data[7], data[8], data[9])
        self.euler_angles = EulerAngles(data[10], data[11], data[12])
        self.quaternion = Quaternion(data[13], data[14], data[15], data[16])
        self.linear_acceleration = LinearAcceleration(
            data[17], data[18], data[19])
        self.gravity_vector = LinearAcceleration(data[20], data[21], data[22])
        self.temperature = data[23]

    def load(self, data, convert=1):
        self.data = data
        if convert == 1:
            self.pressure = float(data[0]) / POZYX_PRESS_DIV_PA
            self.temperature = int(data[23] / POZYX_TEMP_DIV_CELSIUS)
        else:
            self.pressure = float(data[0])
            self.temperature = int(data[23])

        self.acceleration.load(data[1:4], convert)
        self.magnetic.load(data[4:7], convert)
        self.angular_vel.load(data[7:10], convert)
        self.euler_angles.load(data[10:13], convert)
        self.quaternion.load(data[13:17], convert)
        self.linear_acceleration.load(data[17:20], convert)
        self.gravity_vector.load(data[20:23], convert)

    def update_data(self):
        """Not used so data remains the raw unformatted data"""
        pass


class RawSensorData(SensorData):
    """Container for raw sensor data

    This includes, in order, with respective structure:
        - pressure : float
        - acceleration : Acceleration
        - magnetic : Magnetic
        - angular_vel : AngularVelocity
        - euler_angles : EulerAngles
        - quaternion : Quaternion
        - linear_acceleration: LinearAcceleration
        - gravity_vector: LinearAcceleration
        - temperature: Int8
    """

    def __init__(self, data=[0] * 24):
        """Initializes the RawSensorData object"""
        XYZ.__init__(self, data)

    def load(self, data, convert=0):
        SensorData.load(self, data, convert=0)



#!/usr/bin/env python
"""
pypozyx.definitions.registers - contains all register definitions used in Pozyx.

It shouldn't be necessary to use these in basic applications as the library functions
should provide a lot of functionality already, but for advanced users looking to
implement their own low-level functionality, these might be very useful.
"""

# Status registers
POZYX_WHO_AM_I = 0x0   # Returns the constant value 0x43.
POZYX_FIRMWARE_VER = 0x1   # Returns the POZYX firmware version.
POZYX_HARDWARE_VER = 0x2   # Returns the POZYX hardware version.
POZYX_ST_RESULT = 0x3   # Returns the self-test result
POZYX_ERRORCODE = 0x4   # Describes a possibly system error.
POZYX_INT_STATUS = 0x5   # Indicates the source of the interrupt.
POZYX_CALIB_STATUS = 0x6   # Returns the calibration status.

# Configuration registers
POZYX_INT_MASK = 0x10    # Indicates which interrupts are enabled.
POZYX_INT_CONFIG = 0x11  # Configure the interrupt pin
POZYX_CONFIG_LEDS = 0x15    # Configure the LEDs
POZYX_POS_FILTER = 0x14    # Filter used for positioning
POZYX_POS_ALG = 0x16    # Algorithm used for positioning
# Configure the number of anchors and selection procedure
POZYX_POS_NUM_ANCHORS = 0x17
# Defines the update interval in ms in continuous positioning.
POZYX_POS_INTERVAL = 0x18
POZYX_NETWORK_ID = 0x1A    # The network id.
POZYX_UWB_CHANNEL = 0x1C    # UWB channel number.
# Configure the UWB datarate and pulse repetition frequency (PRF)
POZYX_UWB_RATES = 0x1D
POZYX_UWB_PLEN = 0x1E    # Configure the UWB preamble length.
POZYX_UWB_GAIN = 0x1F    # Configure the power gain for the UWB transmitter
POZYX_UWB_XTALTRIM = 0x20    # Trimming value for the uwb crystal.
POZYX_RANGE_PROTOCOL = 0x21    # The ranging protocol
# Configure the mode of operation of the pozyx device
POZYX_OPERATION_MODE = 0x22
POZYX_SENSORS_MODE = 0x23    # Configure the mode of operation of the sensors
POZYX_CONFIG_GPIO1 = 0x27    # Configure GPIO pin 1.
POZYX_CONFIG_GPIO2 = 0x28    # Configure GPIO pin 2.
POZYX_CONFIG_GPIO3 = 0x29    # Configure GPIO pin 3.
POZYX_CONFIG_GPIO4 = 0x2A    # Configure GPIO pin 4.

# Positioning data
POZYX_POS_X = 0x30    # x-coordinate of the device in mm.
POZYX_POS_Y = 0x34    # y-coordinate of the device in mm.
POZYX_POS_Z = 0x38    # z-coordinate of the device in mm.
POZYX_POS_ERR_X = 0x3C    # estimated error covariance of x
POZYX_POS_ERR_Y = 0x3E    # estimated error covariance of y
POZYX_POS_ERR_Z = 0x40    # estimated error covariance of z
POZYX_POS_ERR_XY = 0x42    # estimated covariance of xy
POZYX_POS_ERR_XZ = 0x44    # estimated covariance of xz
POZYX_POS_ERR_YZ = 0x46    # estimated covariance of yz

# Sensor data
POZYX_MAX_LIN_ACC = 0x4E  # Return the max linear acceleration and reset it to 0
POZYX_PRESSURE = 0x50    # Pressure data in mPa
POZYX_ACCEL_X = 0x54    # Accelerometer data (in mg)
POZYX_ACCEL_Y = 0x56
POZYX_ACCEL_Z = 0x58
POZYX_MAGN_X = 0x5A    # Magnemtometer data
POZYX_MAGN_Y = 0x5C
POZYX_MAGN_Z = 0x5E
POZYX_GYRO_X = 0x60    # Gyroscope data
POZYX_GYRO_Y = 0x62
POZYX_GYRO_Z = 0x64
# Euler angles heading (or yaw)  (1 degree = 16 LSB )
POZYX_EUL_HEADING = 0x66
POZYX_EUL_ROLL = 0x68    # Euler angles roll ( 1 degree = 16 LSB )
POZYX_EUL_PITCH = 0x6A    # Euler angles pitch ( 1 degree = 16 LSB )
POZYX_QUAT_W = 0x6C    # Weight of quaternion.
POZYX_QUAT_X = 0x6E    # x of quaternion
POZYX_QUAT_Y = 0x70    # y of quaternion
POZYX_QUAT_Z = 0x72    # z of quaternion
POZYX_LIA_X = 0x74    # Linear acceleration in x-direction
POZYX_LIA_Y = 0x76    # Linear acceleration in y-direction
POZYX_LIA_Z = 0x78    # Linear acceleration in z-direction
POZYX_GRAV_X = 0x7A    # x-component of gravity vector
POZYX_GRAV_Y = 0x7C    # y-component of gravity vector
POZYX_GRAV_Z = 0x7E    # z-component of gravity vector
POZYX_TEMPERATURE = 0x80    # Temperature

# General data
# Returns the number of devices stored internally
POZYX_DEVICE_LIST_SIZE = 0x81
POZYX_RX_NETWORK_ID = 0x82    # The network id of the latest received message
POZYX_RX_DATA_LEN = 0x84    # The length of the latest received message
POZYX_GPIO1 = 0x85    # Value of the GPIO pin 1
POZYX_GPIO2 = 0x86    # Value of the GPIO pin 2
POZYX_GPIO3 = 0x87    # Value of the GPIO pin 3
POZYX_GPIO4 = 0x88    # Value of the GPIO pin 4

# Functions
POZYX_RESET_SYS = 0xB0    # Reset the Pozyx device
POZYX_LED_CTRL = 0xB1    # Control LEDS 1 to 4 on the board
POZYX_TX_DATA = 0xB2    # Write data in the UWB transmit (TX) buffer
POZYX_TX_SEND = 0xB3    # Transmit the TX buffer to some other pozyx device
POZYX_RX_DATA = 0xB4    # Read data from the UWB receive (RX) buffer
POZYX_DO_RANGING = 0xB5    # Initiate ranging measurement
POZYX_DO_POSITIONING = 0xB6    # Initiate the positioning process.
# Set the list of anchor ID's used for positioning.
POZYX_POS_SET_ANCHOR_IDS = 0xB7
# Read the list of anchor ID's used for positioning.
POZYX_POS_GET_ANCHOR_IDS = 0xB8
POZYX_FLASH_RESET = 0xB9  # Reset a portion of the configuration in flash memory
POZYX_FLASH_SAVE = 0xBA  # Store a portion of the configuration in flash memory
POZYX_FLASH_DETAILS = 0xBB  # Return information on what is stored in flash

# Device list functions
# Get all the network IDs's of devices in the device list.
POZYX_DEVICES_GETIDS = 0xC0
# Obtain the network ID's of all pozyx devices within range.
POZYX_DEVICES_DISCOVER = 0xC1
# Obtain the coordinates of the pozyx (anchor) devices within range.
POZYX_DEVICES_CALIBRATE = 0xC2
POZYX_DEVICES_CLEAR = 0xC3    # Clear the list of all pozyx devices.
POZYX_DEVICE_ADD = 0xC4    # Add a pozyx device to the devices list
# Get the stored device information for a given pozyx device
POZYX_DEVICE_GETINFO = 0xC5
# Get the stored coordinates of a given pozyx device
POZYX_DEVICE_GETCOORDS = 0xC6
# Get the stored range inforamation of a given pozyx device
POZYX_DEVICE_GETRANGEINFO = 0xC7
POZYX_CIR_DATA = 0xC8    # Get the channel impulse response (CIR) coefficients



#!/usr/bin/env python
"""pypozyx.definitions.constants - contains all Pozyx constants, such as error definitions, delays, physical convertions."""

POZYX_ANCHOR = 0x00
POZYX_TAG = 0x20

MAX_BUF_SIZE = 100
MAX_SERIAL_SIZE = 28

POZYX_INT_MASK_ALL = 0x1F
POZYX_DELAY_POLLING = 0.002
POZYX_DELAY_LOCAL_WRITE = 0.001
POZYX_DELAY_LOCAL_FUNCTION = 0.005
POZYX_DELAY_REMOTE_WRITE = 0.005
POZYX_DELAY_REMOTE_FUNCTION = 0.01
POZYX_DELAY_INTERRUPT = 0.1
POZYX_DELAY_CALIBRATION = 1
POZYX_DELAY_MODE_CHANGE = 0.02
POZYX_DELAY_RANGING = 0.025
POZYX_DELAY_REMOTE_RANGING = 0.1
POZYX_DELAY_POSITIONING = 0.2
POZYX_DELAY_REMOTE_POSITIONING = 0.4
POZYX_DELAY_FLASH = 0.5
POZYX_FAILURE = 0x0
POZYX_SUCCESS = 0x1
POZYX_TIMEOUT = 0x8
POZYX_3D = 3
POZYX_2D = 2
POZYX_2_5D = 1
POZYX_INT_PIN0 = 0x0
POZYX_INT_PIN1 = 0x1

POZYX_LED_CTRL_LEDRX = 0x10
POZYX_LED_CTRL_LEDTX = 0x20
POZYX_LED_ON = True
POZYX_LED_OFF = False

POZYX_ANCHOR_MODE = 0
POZYX_TAG_MODE = 1

# The GPIO modes
POZYX_GPIO_DIGITAL_INPUT = 0
POZYX_GPIO_PUSHPULL = 1
POZYX_GPIO_OPENDRAIN = 1

# The GPIO pull resistor configuration
POZYX_GPIO_NOPULL = 0
POZYX_GPIO_PULLUP = 1
POZYX_GPIO_PULLDOWN = 2

# anchor selection modes
POZYX_ANCHOR_SEL_MANUAL = 0
POZYX_ANCHOR_SEL_AUTO = 1

# discovery options
POZYX_DISCOVERY_ANCHORS_ONLY = 0
POZYX_DISCOVERY_TAGS_ONLY = 1
POZYX_DISCOVERY_ALL_DEVICES = 2

# positioning algorithm options
POZYX_POS_ALG_UWB_ONLY = 0
POZYX_POS_ALG_TRACKING = 4

# ranging protocol options
POZYX_RANGE_PROTOCOL_PRECISION = 0x00
POZYX_RANGE_PROTOCOL_FAST = 0x01
POZYX_RANGE_PROTOCOL_TEST = 0x02

# positioning filters
FILTER_TYPE_NONE = 0
FILTER_TYPE_FIR = 1
FILTER_TYPE_MOVINGAVERAGE = 3
FILTER_TYPE_MOVINGMEDIAN = 4

# how to intercept pozyx events: by polling or by interrupts
MODE_POLLING = 0
MODE_INTERRUPT = 1

# Division factors for converting the raw register values to meaningful
# physical quantities
POZYX_POS_DIV_MM = 1.0
POZYX_PRESS_DIV_PA = 1000.0
POZYX_ACCEL_DIV_MG = 1.0
POZYX_GYRO_DIV_DPS = 16.0
POZYX_MAG_DIV_UT = 16.0
POZYX_EULER_DIV_DEG = 16.0
POZYX_QUAT_DIV = 16384.0
POZYX_TEMP_DIV_CELSIUS = 1.0

# error-code definitions
POZYX_ERROR_NONE = 0x00
POZYX_ERROR_I2C_WRITE = 0x01
POZYX_ERROR_I2C_CMDFULL = 0x02
POZYX_ERROR_ANCHOR_ADD = 0x03
POZYX_ERROR_COMM_QUEUE_FULL = 0x04
POZYX_ERROR_I2C_READ = 0x05
POZYX_ERROR_UWB_CONFIG = 0x06
POZYX_ERROR_OPERATION_QUEUE_FULL = 0x07
POZYX_ERROR_TDMA = 0xA0
POZYX_ERROR_STARTUP_BUSFAULT = 0x08
POZYX_ERROR_FLASH_INVALID = 0x09
POZYX_ERROR_NOT_ENOUGH_ANCHORS = 0x0A
POZYX_ERROR_DISCOVERY = 0X0B
POZYX_ERROR_CALIBRATION = 0x0C
POZYX_ERROR_FUNC_PARAM = 0x0D
POZYX_ERROR_ANCHOR_NOT_FOUND = 0x0E
POZYX_ERROR_FLASH = 0x0F
POZYX_ERROR_MEMORY = 0x10
POZYX_ERROR_RANGING = 0x11
POZYX_ERROR_RTIMEOUT1 = 0x12
POZYX_ERROR_RTIMEOUT2 = 0x13
POZYX_ERROR_TXLATE = 0x14
POZYX_ERROR_UWB_BUSY = 0x15
POZYX_ERROR_POSALG = 0x16
POZYX_ERROR_NOACK = 0x17
POZYX_ERROR_SNIFF_OVERFLOW = 0xE0
POZYX_ERROR_NO_PPS = 0xF0
POZYX_ERROR_NEW_TASK = 0xF1
POZYX_ERROR_UNRECDEV = 0xFE
POZYX_ERROR_GENERAL = 0xFF

# flash configuration types
POZYX_FLASH_REGS = 1
POZYX_FLASH_ANCHOR_IDS = 2
POZYX_FLASH_NETWORK = 3
POZYX_FLASH_ALL = 4

# possible pin configuration settings
POZYX_INT_CONFIG = 0x24
PIN_MODE_PUSHPULL = 0
PIN_MODE_OPENDRAIN = 1

PIN_ACTIVE_LOW = 0
PIN_ACTIVE_HIGH = 1

POZYX_ALL_CHANNELS = [1, 2, 3, 4, 5, 7]
POZYX_ALL_BITRATES = [0, 1, 2]
POZYX_ALL_PRFS = [1, 2]
POZYX_ALL_PLENS = [0x04, 0x14, 0x24, 0x34, 0x08, 0x18, 0x28, 0x0C]



#!/usr/bin/env python
"""
pypozyx

PyPozyx
=======

Warning: This library isn't considered finished, please mail to
support@pozyx.io or create a GitHub issue if you experience an issue.

Provides:
    - A way to interact with Pozyx over USB through PozyxSerial
    - A solid set of easy-to-use and easy-to-expand functions for
      working with Pozyx.

NB: At this point, documentation is planned for functions and the structures
relevant for users. This constitutes the lib, structures.device, and
structures.sensor_data modules.

How to use the documentation
----------------------------
Documentation is available in two forms: docstrings provided with the code,
and the documentation available on `the Pozyx documentation <https://www.pozyx.io/Documentation/Datasheet/arduino>`_.
For now, this links to the Arduino library functions, as there is a lot of
similarity and evidently both libraries influence each other. If you're not
working low-level, you shouldn't notice much difference between the two.
When the Python webpage comes only, we'll clearly distinguish between
functions not available in the Arduino lib.

The docstring examples assume that, for ease of use and typing, everything
from the pypozyx package has been imported using::

  >>> from pypozyx import *

While not considered good practice, it makes reading and writing scripts
much easier when starting out.

This imports PozyxSerial, all constants definitions, and the device, generic,
and sensordata structures. This does not import the registers definitions,
or the bitmasks, as these are considered advanced features and assume the
user knows what he's doing. To access these, import the modules
pypozyx.definitions.bitmasks and pypozyx.definitions.registers.

Whenever you see the following code in the docstrings::

  >>> pozyx.anyFunction()
  >>> # or
  >>> p.anyFunction()

'pozyx' and 'p' are the actual Pozyx device's interface instance, not a module.
This means that the Pozyx interface instance performs anyFunction.

Available subpackages
---------------------
structures
    Basic and specialised data structures tailored to using Pozyx
definitions
    Pozyx constant definitions, such as general constants, register indexes,
    and bit masks for certain functionality.

State of the beta
-----------------
The PyPozyx package is currently in Beta version 1. In this Beta, the initial
run of documentation will be added, and bugfixes will be done as encountered.
When this is done, the library will enter Beta version 2, which is the final
planned beta. In this, we will focus on robustness and do a final documentation
evaluation. After this, we'll consider it 1.0.

Serial port warnings
--------------------
Make  sure that the serial connection is closed (i.e. program terminated)
before disconnecting Pozyx. Terminating a running program can be done using
Ctrl+Z or Ctrl+C. If something goes wrong, chances are high you will have to
edit which serial port is used by Pozyx, as the previous one will be
considered unavailable by your OS.

Automatically selecting serial port
-----------------------------------
A trick to remedy this is by using the following code::

  >>> import serial.tools.list_ports
  >>> port = list(serial.tools.list_ports.comports())[0]

This only works consistently when Pozyx is the only serial device you're
working with, when you use more devices, their order in the list may
switch at some point.

"""

from pypozyx.definitions.constants import *
from pypozyx.pozyx_serial import (PozyxSerial, get_pozyx_ports,
                                  get_serial_ports, list_serial_ports)
from pypozyx.structures.device import *
from pypozyx.structures.generic import *
from pypozyx.structures.sensor_data import *



# A first test fixture trial.

from pypozyx import *

import unittest

port = '/dev/ttyACM0'


class PozyxSerialCoreTest(unittest.TestCase):

    def setUp(self):
        self.pozyx = PozyxSerial(port)

    def tearDown(self):
        self.pozyx.ser.close()

    def test_who_am_i(self):
        whoami = Data([0])
        self.pozyx.getWhoAmI(whoami)

        # WhoAmI register should always return 0x43
        self.assertEqual(0x43, whoami[0], msg="WhoAmI!= 0x43")
        #print("WhoAmI: %0.2x" % whoami[0])

    def test_network_id(self):
        network_id = NetworkID(0)
        self.pozyx.getNetworkId(network_id)

        self.assertLess(0, network_id[0])
        print("Network ID: ", network_id)


if __name__ == "__main__":
    unittest.main(warnings='ignore')



#!/usr/bin/env python
"""
pypozyx.structures.generic - introduces generic data structures derived from ByteStructure

Generic Structures
==================

As the name implies, contains generic structures whose specific use is up to the
user. You should use SingleRegister where applicable when reading/writing
a single register, and use Data for larger data structures.

Structures contained
--------------------
Data
    THE generic data structure, a powerful way of constructing arbitrarily
    formed packed data structures
XYZ
    A generic XYZ data structure that is used in much 3D sensor data
SingleRegister
    Data resembling a single register. Can choose size and whether signed.
UniformData
    A variation on Data with all data being a uniform format. Questionably useful.

The use of Data
---------------
Data creates a packed data structure with size and format that is entirely the user's choice.
The format follows the one used in struct, where b is a byte, h is a 2-byte int, and
i is a default-sized integer, and f is a float. In capitals, these are signed.
So, to create a custom construct consisting of 4 uint16 and a single int, the
following code can be used.

  >>> d = Data([0] * 5, 'HHHHi')

or

  >>> data_format = 'HHHHi'
  >>> d = Data([0] * len(data_format), data_format)
"""

import struct

from pypozyx.structures.byte_structure import ByteStructure


def is_reg_readable(reg):
    """Returns whether a Pozyx register is readable."""
    if (reg >= 0x00 and reg < 0x07) or (reg >= 0x10 and reg < 0x12) or (reg >= 0x14 and reg < 0x22) or (reg >= 0x22 and reg < 0x24) or (reg >= 0x27 and reg < 0x2B) or (reg >= 0x30 and reg < 0x48) or (reg >= 0x4E and reg < 0x89):
        return True
    return False


def is_reg_writable(reg):
    """Returns whether a Pozyx register is writeable."""
    if (reg >= 0x10 and reg < 0x12) or (reg >= 0x14 and reg < 0x22) or (reg >= 0x22 and reg < 0x24) or (reg >= 0x27 and reg < 0x2B) or (reg >= 0x30 and reg < 0x3C) or (reg >= 0x85 and reg < 0x89):
        return True
    return False


def is_functioncall(reg):
    """Returns whether a Pozyx register is a Pozyx function."""
    if (reg >= 0xB0 and reg < 0xBC) or (reg >= 0xC0 and reg < 0xC9):
        return True
    return False


def dataCheck(data):
    """Returns whether an object is part of the ByteStructure-derived classes or not.

    The function checks the base classes of the passed data object. This function enables
    many library functions to be passed along its data as either an int/list or the properly
    intended data structure. For example, the following code will result in the
    same behaviour::

      >>> p.setCoordinates([0, 0, 0])
      >>> # or
      >>> coords =  Coordinates()
      >>> p.setCoordinates(coords)

    AND

      >>> p.setNetworkId(0x6000)
      >>> # or
      >>> n = NetworkID(0x6000)
      >>> p.setNetworkId(n)

    Note that this only works for functions where you change one of the Pozyx's
    settings. When reading data from the Pozyx, you have to pass along the correct
    data structure.

    Using dataCheck
    ---------------
    You might want to use this in your own function, as it makes it more robust
    to whether an int or list gets sent as a parameter to your function, or a
    ByteStructure-like object. If so, you can perform::

      >>> if not dataCheck(sample): # assume a is an int but you want it to be a SingleRegister
      >>>     sample = SingleRegister(sample)

    """
    if not(Data in type(data).__bases__ or ByteStructure in type(data).__bases__ or Data is type(data) or XYZ in type(data).__bases__):
        return False
    return True


class XYZ(ByteStructure):
    """
    Generic XYZ data structure consisting of 3 integers x, y, and z.

    Not recommended to use in practice, as relevant sensor data classes are derived from this.

    If deriving this, don't forget to implement your own update_data function, or data will
    be [x, y, z] consistently instead of [..., x, y, z, ...].
    """
    physical_convert = 1

    byte_size = 12
    data_format = 'iii'

    def __init__(self, x=0, y=0, z=0):
        """Initializes the XYZ or XYZ-derived object."""
        self.x = x
        self.y = y
        self.z = z
        self.data = [x, y, z]

    def load(self, data=[0] * 3, convert=1):
        self.data = data
        if convert:
            self.x = data[0] / self.physical_convert
            self.y = data[1] / self.physical_convert
            self.z = data[2] / self.physical_convert
        else:
            self.x = data[0]
            self.y = data[1]
            self.z = data[2]

    def update_data(self):
        try:
            if self.data != [self.x, self.y, self.z]:
                self.data = [self.x, self.y, self.z]
        except:
            return

    def __str__(self):
        return 'X: {self.x}, Y: {self.y}, Z: {self.z}'.format(self=self)


class Data(ByteStructure):
    """Data allows the user to define arbitrary data structures to use with Pozyx.

    The Leatherman of ByteStructure-derived classes, Data allows you to create your own
    library-compatible packed data structures. Also for empty data, this is used.

    The use of Data
    ---------------
    Data creates a packed data structure with size and format that is entirely the user's choice.
    The format follows the one used in struct, where b is a byte, h is a 2-byte int, and
    i is a default-sized integer, and f is a float. In capitals, these are signed.
    So, to create a custom construct consisting of 4 uint16 and a single int, the
    following code can be used.

      >>> d = Data([0] * 5, 'HHHHi')

    or

      >>> data_format = 'HHHHi'
      >>> d = Data([0] * len(data_format), data_format)

    Kwargs:
        data: Data contained in the data structure. When no data_format is passed, these are assumed UInt8 values.
        data_format: Custom data format for the data passed.
    """

    def __init__(self, data=None, data_format=None):
        if data is None:
            data = []
        self.data = data
        if data_format is None:
            data_format = 'B' * len(data)
        self.data_format = data_format
        self.set_packed_size()
        self.byte_data = '00' * self.byte_size

    def load(self, data, convert=1):
        self.data = data


class SingleRegister(Data):
    """
    SingleRegister is container for the data from a single Pozyx register.

    By default, this represents a UInt8 register. Used for both reading and writing.
    The size and whether the data is a 'signed' integer are both changeable by the
    user using the size and signed keyword arguments.

    Kwargs:
        value: Value of the register.
        size: Size of the register. 1, 2, or 4. Default 0.
        signed: Whether the data is signed. unsigned by default.
        print_hex: How to print the register output. Hex by default. Special options are 'hex' and 'bin'
            other things, such as 'dec', will return decimal output.
    """

    def __init__(self, value=0, size=1, signed=0, print_style='hex'):
        self.print_style = print_style
        if size == 1:
            data_format = 'b'
        elif size == 2:
            data_format = 'h'
        elif size == 4:
            data_format = 'i'
        if signed == 0:
            data_format = data_format.capitalize()
        Data.__init__(self, [value], data_format)

    def load(self, data, convert=1):
        self.data = data

    @property
    def value(self):
        return self.data[0]

    @value.setter
    def value(self, new_value):
        self.data[0] = new_value


    def __str__(self):
        if self.print_style is 'hex':
            return hex(self.data[0]).capitalize()
        elif self.print_style is 'bin':
            return bin(self.data[0])
        else:
            return str(self.data[0])



# -*- coding: utf-8 -*-
from distutils.core import setup

setup(
    name='pypozyx',
    packages=['pypozyx', 'pypozyx.definitions',
              'pypozyx.structures', 'pypozyx.tests'],
    version='1.0.0',
    description='Python library for Pozyx',
    author='Laurent Van Acker',
    author_email='laurent@pozyx.io',
    url = 'https://github.com/pozyxLabs/Pozyx-Python-library',
    download_url='https://github.com/pozyxLabs/Pozyx-Python-library/archive/v1.0.tar.gz',
    keywords=['pozyx', 'serial', 'positioning', 'localisation'],
    classifiers=[
        'Programming Language :: Python',
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: End Users/Desktop',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Scientific/Engineering',
    ],
    long_description="""\
    Pozyx Python library, used to interface with the Pozyx without need of an arduino.
    Currently supports USB (serial), planned I2C support.

    Both Python 2 and 3 are supported.
    """
)



#!/usr/bin/env python
"""pypozyx.core - core Pozyx interface and inter-Pozyx communication functionality through the PozyxCore class"""
from time import sleep, time

from pypozyx.definitions.bitmasks import (POZYX_INT_STATUS_ERR,
                                          POZYX_INT_STATUS_FUNC)
from pypozyx.definitions.constants import (MAX_BUF_SIZE, MAX_SERIAL_SIZE,
                                           POZYX_DELAY_LOCAL_WRITE,
                                           POZYX_DELAY_POLLING,
                                           POZYX_DELAY_REMOTE_WRITE,
                                           POZYX_FAILURE, POZYX_SUCCESS,
                                           POZYX_TIMEOUT)
from pypozyx.definitions.registers import (POZYX_INT_STATUS, POZYX_RX_DATA,
                                           POZYX_RX_NETWORK_ID, POZYX_TX_DATA,
                                           POZYX_TX_SEND)
from pypozyx.structures.generic import (Data, SingleRegister, dataCheck,
                                        is_functioncall, is_reg_readable,
                                        is_reg_writable)


class PozyxCore():
    """PozyxCore

    PozyxCore
    =========

    Implements virtual core Pozyx interfacing functions such as regRead,
    regWrite and regFunction, which have to be implemented in the derived interface.
    Auxiliary functions for core functionality, getRead, setWrite, useFunction,
    and checkForFlag, are also included in PozyxCore.

    waitForFlag_safe, which uses polling, is also implemented interface-independently.

    Apart from these core interface functions, core inter-Pozyx communication
    functionality is also implemented here. This includes remote interface functions
    and general communication to send and receive data from other devices.
    """

    def __init__(self):
        """
        Constructor for PozyxCore.

        PozyxCore isn't practically usable on its own as it misses a core interface
        implementation.
        """
        pass

    def regRead(self, address, data):
        """
        Stores the read amount of bytes equal to data's size starting at address into data.

        This is a virtual function, be sure to implement this in your own interface.
        """
        raise NotImplementedError(
            'You need to override this function in your derived interface!')

    def regWrite(self, address, data):
        """
        Writes the given data starting at address.

        This is a virtual function, be sure to implement this in your own interface.
        """
        raise NotImplementedError(
            'You need to override this function in your derived interface!')

    def regFunction(self, address, params, data):
        """
        Performs a function with given parameters, storing its output in data.

        This is a virtual function, be sure to implement this in your own interface.
        """
        raise NotImplementedError(
            'You need to override this function in your derived interface!')

    def remoteRegWrite(self, destination, address, data):
        """
        Performs regWrite on a remote Pozyx device.

        Args:
            destination: Network ID of destination device. integer ID or NetworkID(ID).
            address: Register address to start the writing operation on.
            data: Contains the data to be written. ByteStructure-derived object.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        if is_reg_writable(address) == 0:
            return POZYX_FAILURE
        if len(data) > MAX_BUF_SIZE - 1:
            return POZYX_FAILURE

        send_data = Data([0, address] + data.data, 'BB' + data.data_format)
        status = self.regFunction(POZYX_TX_DATA, send_data, Data([]))
        if status == POZYX_FAILURE:
            return status

        self.regRead(POZYX_INT_STATUS, SingleRegister())
        params = Data([destination, 0x04], 'HB')
        status = self.regFunction(POZYX_TX_SEND, params, Data([]))
        if status == POZYX_FAILURE:
            return status
        return self.checkForFlag(POZYX_INT_STATUS_FUNC, 0.1)

    def remoteRegRead(self, destination, address, data):
        """
        Performs regRead on a remote Pozyx device.

        Args:
            destination: Network ID of destination device. integer ID or NetworkID(ID).
            address: Register address to start the read operation from.
            data: Container for the read data. ByteStructure-derived object.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        if dataCheck(destination):
            destination = destination[0]
        if is_reg_readable(address) == 0:
            return POZYX_FAILURE
        if len(data) > MAX_BUF_SIZE:
            return POZYX_FAILURE
        if destination == 0:
            return POZYX_FAILURE

        send_data = Data([0, address, data.byte_size])
        status = self.regFunction(POZYX_TX_DATA, send_data, Data([]))
        if status == POZYX_FAILURE:
            return status

        self.regRead(POZYX_INT_STATUS, SingleRegister())
        params = Data([destination, 0x02], 'HB')
        status = self.regFunction(POZYX_TX_SEND, params, Data([]))
        if status == POZYX_FAILURE:
            return status

        status = self.checkForFlag(POZYX_INT_STATUS_FUNC, 1)
        if status == POZYX_SUCCESS:
            rx_info = Data([0, 0], 'HB')
            self.regRead(POZYX_RX_NETWORK_ID, rx_info)
            if rx_info[0] == destination and rx_info[1] == data.byte_size:
                status = self.readRXBufferData(data)
                return status
            else:
                return POZYX_FAILURE
        return status

    def remoteRegFunction(self, destination, address, params, data):
        """
        Performs regFunction on a remote Pozyx device.

        Args:
            destination: Network ID of destination device. integer ID or NetworkID(ID).
            address: Register address to start the read operation from.
            params: Parameters for the register function. ByteStructure-derived object of uint8s.
            data: Container for the data returned by the register function. ByteStructure-derived object.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        if is_functioncall(address) == 0:
            return POZYX_FAILURE

        send_data = Data([0, address] + params.data, 'BB' + params.data_format)
        status = self.regFunction(POZYX_TX_DATA, send_data, Data([]))
        if status == POZYX_FAILURE:
            return status

        self.regRead(POZYX_INT_STATUS, SingleRegister())
        tx_params = Data([destination, 0x08], 'HB')
        status = self.regFunction(POZYX_TX_SEND, tx_params, Data([]))
        if status == POZYX_FAILURE:
            return status

        status = self.checkForFlag(POZYX_INT_STATUS_FUNC, 1)
        if status == POZYX_SUCCESS:
            rx_info = Data([0, 0], 'HB')
            self.regRead(POZYX_RX_NETWORK_ID, rx_info)
            if rx_info[0] == destination and rx_info[1] == data.byte_size + 1:
                return_data = Data([0] + data.data, 'B' + data.data_format)
                status = self.readRXBufferData(return_data)
                if status == POZYX_FAILURE:
                    return status
                if len(return_data) > 1:
                    data.load(return_data[1:])
                return return_data[0]
            else:
                return POZYX_FAILURE
        return status

    def waitForFlag(self, interrupt_flag, timeout_s, interrupt=None):
        """
        Checks the interrupt register for given flag until encountered/past the timeout time.

        This is a virtual function, be sure to implement this in your derived interface.
        """
        raise NotImplementedError(
            'You need to override this function in your derived interface!')

    def waitForFlag_safe(self, interrupt_flag, timeout_s, interrupt=None):
        """
        Performs waitForFlag in polling mode.

        Args:
            interrupt_flag: Flag of interrupt type to check the interrupt register against.
            timeout_s: duration to wait for the interrupt in seconds.

        Kwags:
            interrupt: Container for the interrupt status register data.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        if interrupt is None:
            interrupt = SingleRegister()
        start = time()
        while(time() - start < timeout_s):
            sleep(POZYX_DELAY_POLLING)
            status = self.regRead(POZYX_INT_STATUS, interrupt)
            if (interrupt[0] & interrupt_flag) and status == POZYX_SUCCESS:
                return True
        return False

    ## \addtogroup core
    # @{

    def setWrite(self, address, data, remote_id=None, local_delay=POZYX_DELAY_LOCAL_WRITE, remote_delay=POZYX_DELAY_REMOTE_WRITE):
        """
        Writes data to Pozyx registers either locally or remotely.

        Args:
            address: The register address
            data: A ByteStructure - derived object that contains the data to be written.

        Kwargs:
            remote_id: Remote ID for remote read.
            local_delay: Delay after a local write
            remote_delay: Delay after a remote write

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT

        Examples:
            >>> leds = SingleRegister(0xFF)
            >>> self.setWrite(POZYX_LED_CTRL, leds)
        """
        if remote_id is None:
            status = self.regWrite(address, data)
            sleep(local_delay)
        else:
            status = self.remoteRegWrite(remote_id, address, data)
            sleep(remote_delay)
        return status

    def getRead(self, address, data, remote_id=None):
        """
        Reads Pozyx register data either locally or remotely.

        Args:
            address: The register address
            data: A ByteStructure - derived object that is the container of the read data.

        Kwargs:
            remote_id: Remote ID for remote read.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT

        Example:
            >>> whoami = SingleRegister()
            >>> self.getRead(POZYX_WHO_AM_I, whoami)
            >>> print(whoami)
            67
        """
        if remote_id is None:
            return self.regRead(address, data)
        else:
            return self.remoteRegRead(remote_id, address, data)

    def useFunction(self, function, params=None, data=None, remote_id=None):
        """
        Activates a Pozyx register function either locally or remotely.

        Args:
            address: The function address

        Kwargs:
            params: A ByteStructure - derived object that contains the parameters for the function.
            data: A ByteStructure - derived object that is the container of the read data.
            remote_id: Remote ID for remote read.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT

        Example:
            >>> self.useFunction(POZYX_DEVICES_CLEAR)
        """
        if params is None:
            params = Data([])
        if data is None:
            data = Data([])
        if remote_id is None:
            status = self.regFunction(function, params, data)
        else:
            status = self.remoteRegFunction(
                remote_id, function, params, data)
        return status

    # wait for flag functions

    def checkForFlag(self, interrupt_flag, timeout_s, interrupt=None):
        """Performs waitForFlag_safe and checks against errors or timeouts.

        This abstracts the waitForFlag status check routine commonly encountered
        in more complex library functions and checks the given flag against
        the error flag.

        Args:
            interrupt_flag: Flag of interrupt type to check the interrupt register against.
            timeout_s: duration to wait for the interrupt in seconds

        Kwags:
            interrupt: Container for the interrupt status register data.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        if interrupt is None:
            interrupt = SingleRegister()
        if self.waitForFlag_safe(interrupt_flag | POZYX_INT_STATUS_ERR, timeout_s, interrupt):
            if (interrupt[0] & POZYX_INT_STATUS_ERR) == POZYX_INT_STATUS_ERR:
                return POZYX_FAILURE
            else:
                return POZYX_SUCCESS
        else:
            return POZYX_TIMEOUT

    ## @}

    ## \addtogroup communication_functions

    def readRXBufferData(self, data):
        """
        Reads the device's receive buffer's data completely.

        Args:
            data: Container for the data to be read from the receiver buffer.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE
        """
        if data.byte_size > MAX_BUF_SIZE:
            return POZYX_FAILURE

        status = POZYX_SUCCESS
        if data.byte_size < MAX_SERIAL_SIZE:
            params = Data([0, data.byte_size])
            status &= self.regFunction(POZYX_RX_DATA, params, data)
        else:
            runs = int(data.byte_size / MAX_SERIAL_SIZE)
            s = ''
            for i in range(runs):
                params = Data([i * MAX_SERIAL_SIZE, MAX_SERIAL_SIZE])
                d = Data([0] * MAX_SERIAL_SIZE)
                status &= self.regFunction(POZYX_RX_DATA, params, d)
                s += d.byte_data
            params = Data(
                [runs * MAX_SERIAL_SIZE, data.byte_size - runs * MAX_SERIAL_SIZE])
            d = Data([0] * (data.byte_size - runs * MAX_SERIAL_SIZE))
            status &= self.regFunction(POZYX_RX_DATA, params, d)
            s += d.byte_data
            data.load_bytes(s)
        return status

    def writeTXBufferData(self, data, offset=0):
        """
        Writes data to the device's transmit buffer at the offset address.

        Args:
            data: Data to write to the Pozyx buffer. Has to be a ByteStructure derived object.

        Kwargs:
            offset: Offset in buffer to start writing data

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE
        """
        if offset + data.byte_size > MAX_BUF_SIZE:
            return POZYX_FAILURE

        # have to account for the parameter taking up a byte
        _MAX_SERIAL_SIZE = MAX_SERIAL_SIZE - 1

        status = POZYX_SUCCESS
        data = Data(data.transform_to_bytes())
        runs = int(data.byte_size / _MAX_SERIAL_SIZE)
        for i in range(runs):
            params = Data([i * _MAX_SERIAL_SIZE] + data[i *
                                                        _MAX_SERIAL_SIZE: (i + 1) * _MAX_SERIAL_SIZE])
            status &= self.regFunction(POZYX_TX_DATA, params, Data([]))
        params = Data([runs * _MAX_SERIAL_SIZE] +
                      data[runs * _MAX_SERIAL_SIZE:])
        return status & self.regFunction(POZYX_TX_DATA, params, Data([]))

    def sendTXBufferData(self, destination):
        """
        Sends the transmit buffer's data to the destination device.

        Args:
            destination: Network ID of destination. integer ID or NetworkID(ID)

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE
        """
        if dataCheck(destination):
            destination = destination[0]  # see what I did there.
        params = Data([destination, 0x06], 'HB')
        status = self.regFunction(POZYX_TX_SEND, params, Data([]))
        return status

    def sendData(self, destination, data):
        """
        Stores the data in the transmit buffer and then sends it to the device with ID destination.

        Args:
            destination: Network ID of destination. integer ID or NetworkID(ID)
            data: Data to send to the destination. Has to be a ByteStructure derived object.

        Performs the following code::

          >>>self.writeTXBufferData(data)
          >>>self.sendTXBufferData(destination)

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE
        """
        status = POZYX_SUCCESS
        status &= self.writeTXBufferData(data)
        status &= self.sendTXBufferData(destination)
        return status

    ## @}



#!/usr/bin/env python
"""pypozyx.definitions.bitmasks - contains all bitmasks used in Pozyx functionality, such as interrupt flags."""

# Bit mask for POZYX_ST_RESULT
POZYX_ST_RESULT_ACC = 0x01
POZYX_ST_RESULT_MAGN = 0x02
POZYX_ST_RESULT_GYR = 0x04
POZYX_ST_RESULT_MCU = 0x08
POZYX_ST_RESULT_PRES = 0x10
POZYX_ST_RESULT_UWB = 0x20

# Bit mask for POZYX_INT_STATUS
POZYX_INT_STATUS_ERR = 0x01
POZYX_INT_STATUS_POS = 0x02
POZYX_INT_STATUS_IMU = 0x04
POZYX_INT_STATUS_RX_DATA = 0x08
POZYX_INT_STATUS_FUNC = 0x10

# Bit mask for POZYX_INT_MASK
POZYX_INT_MASK_ERR = 0x01
POZYX_INT_MASK_POS = 0x02
POZYX_INT_MASK_IMU = 0x04
POZYX_INT_MASK_RX_DATA = 0x08
POZYX_INT_MASK_FUNC = 0x10
POZYX_INT_MASK_TDMA = 0x40
POZYX_INT_MASK_PIN = 0x80
POZYX_INT_MASK_ALL = 0x1F

# Bit mask for POZYX_LED_CTRL
POZYX_LED_CTRL_LED1 = 0x01
POZYX_LED_CTRL_LED2 = 0x02
POZYX_LED_CTRL_LED3 = 0x04
POZYX_LED_CTRL_LED4 = 0x08

# Bit mask for device type
POZYX_TYPE = 0xE0



"""pypozyx.pozyx_serial - contains the serial interface with Pozyx through PozyxSerial."""
from time import sleep

from pypozyx.definitions.constants import *
from pypozyx.definitions.registers import *
from pypozyx.lib import PozyxLib
from pypozyx.structures.generic import Data, SingleRegister
from serial import Serial
from serial.tools.list_ports import comports

## \addtogroup auxiliary_serial
# @{

def list_serial_ports():
    """Prints the open serial ports line per line"""
    ports = comports()
    for port in ports:
        print(port)


def get_serial_ports():
    """Returns the open serial ports"""
    return comports()


def get_pozyx_ports():
    """Returns the Pozyx serial ports. Untested on UNIX"""
    ports = get_serial_ports()
    pozyx_ports = []
    for port in ports:
        if "STMicroelectronics Virtual COM Port" in port.description:
            pozyx_ports.append(port.device)

## @}


class PozyxSerial(PozyxLib):
    """
    PozyxSerial
    ===========

    This class provides the Pozyx Serial interface, and opens and locks the serial
    port to use with Pozyx. All functionality from PozyxLib and PozyxCore is included.

    Args:
        port: string name of the serial port. On UNIX this will be '/dev/ttyACMX', on
            Windows this will be 'COMX', with X a random number.

    Kwargs:
        baudrate: the baudrate of the serial port. Default value is 115200.
        timeout: timeout for the serial port communication in seconds. Default is 0.1s or 100ms.
        print_output: boolean for debugging purposes. If set to True,

    Example usage:
        >>> pozyx = PozyxSerial('COMX') # Windows
        >>> pozyx = PozyxSerial('/dev/ttyACMX', print_output=True) # Linux and OSX. Also puts debug output on.

    Finding the serial port
    =======================
    Finding the serial port can be easily done with the following code:
        >>> import serial.tools.list_ports
        >>> print serial.tools.list_ports.comports()[0]

    Putting one and two together, automating the correct port selection with one Pozyx attached:
        >>> import serial.tools.list_ports
        >>> pozyx = PozyxSerial(serial.tools.list_ports.comports()[0])
    """

    ## \addtogroup core
    # @{
    def __init__(self, port, baudrate=115200, timeout=0.1, write_timeout=0.1, print_output=False, debug_trace=False):
        """Initializes the PozyxSerial object. See above for details."""
        self.print_output = print_output
        try:
            self.ser = Serial(port, baudrate, timeout=timeout,
                              write_timeout=write_timeout)
        except:
            print(
                "Couldn't connect with Pozyx, wrong/busy serial port, or pySerial not installed.")
            if debug_trace:
                import traceback
                import sys
                traceback.print_tb(sys.exc_info()[2])
            raise SystemExit

        sleep(0.25)

        regs = Data([0, 0, 0])
        if self.regRead(POZYX_WHO_AM_I, regs) == POZYX_FAILURE:
            print("Connected to Pozyx, but couldn't read serial data.")
            if debug_trace:
                import traceback
                import sys
                traceback.print_tb(sys.exc_info()[2])
            raise SystemExit

        self._hw_version = regs[1]
        self._sw_version = regs[2]

        if regs[0] != 0x43:
            print("WHO AM I returned 0x%0.2x, something is wrong with Pozyx." %
                  regs[0])
            raise SystemExit

    ## @}

    def regWrite(self, address, data):
        """
        Writes data to the Pozyx registers, starting at a register address, if registers are writable.

        Args:
            address: Register address to start writing at.
            data: Data to write to the Pozyx registers. Has to be ByteStructure-derived object.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE
        """
        data.load_hex_string()
        index = 0
        runs = int(data.byte_size / MAX_SERIAL_SIZE)
        for i in range(runs):
            s = 'W,%0.2x,%s\r' % (
                address + index, data.byte_data[2 * index: 2 * (index + MAX_SERIAL_SIZE)])
            index += MAX_SERIAL_SIZE
            try:
                self.ser.write(s.encode())
            except:
                return POZYX_FAILURE
            # delay(POZYX_DELAY_LOCAL_WRITE)
        s = 'W,%0.2x,%s\r' % (address + index, data.byte_data[2 * index:])
        try:
            self.ser.write(s.encode())
        except:
            return POZYX_FAILURE
        return POZYX_SUCCESS

    def serialExchange(self, s):
        """
        Auxiliary. Performs a serial write to and read from the Pozyx.

        Args:
            s: Serial message to send to the Pozyx
        Returns:
            Serial message the Pozyx returns, stripped from 'D,' at its start
                and NL+CR at the end.
        """
        self.ser.write(s.encode())
        response = self.ser.readline().decode()
        if self.print_output:
            print('The response to %s is %s.' % (s.strip(), response.strip()))
        if len(response) == 0:
            raise EnvironmentError
        if response[0] == 'D':
            return response[2:-2]
        raise EnvironmentError

    def regRead(self, address, data):
        """
        Reads data from the Pozyx registers, starting at a register address, if registers are readable.

        Args:
            address: Register address to start writing at.
            data: Data to write to the Pozyx registers. Has to be ByteStructure-derived object.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE
        """
        runs = int(data.byte_size / MAX_SERIAL_SIZE)
        r = ''
        for i in range(runs):
            s = 'R,%0.2x,%i\r' % (
                address + i * MAX_SERIAL_SIZE, MAX_SERIAL_SIZE)
            try:
                r += self.serialExchange(s)
            except:
                return POZYX_FAILURE
        s = 'R,%0.2x,%i\r' % (
            address + runs * MAX_SERIAL_SIZE, data.byte_size - runs * MAX_SERIAL_SIZE)
        try:
            r += self.serialExchange(s)
        except:
            return POZYX_FAILURE
        data.load_bytes(r)
        return POZYX_SUCCESS

    def regFunction(self, address, params, data):
        """
        Performs a register function on the Pozyx, if the address is a register function.

        Args:
            address: Register function address of function to perform.
            params: Parameters for the register function. Has to be ByteStructure-derived object.
            data: Container for the data the register function returns. Has to be ByteStructure-derived object.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE
        """
        params.load_hex_string()
        s = 'F,%0.2x,%s,%i\r' % (address, params.byte_data, data.byte_size + 1)
        try:
            r = self.serialExchange(s)
        except:
            return POZYX_FAILURE
        if len(data) > 0:
            data.load_bytes(r[2:])
        return int(r[0:2], 16)

    def waitForFlag(self, interrupt_flag, timeout_s, interrupt=None):
        """
        Waits for a certain interrupt flag to be triggered, indicating that that type of interrupt
        occured.

        Args:
            interrupt_flag: Flag indicating interrupt type.
            timeout_s: time in seconds that POZYX_INT_STATUS will be checked for the flag before returning
                POZYX_TIMEOUT.

        Kwargs:
            interrupt: Container for the POZYX_INT_STATUS data. For debugging purposes.

        Returns:
            POZYX_SUCCESS, POZYX_FAILURE, POZYX_TIMEOUT
        """
        if interrupt is None:
            interrupt = SingleRegister()
        return self.waitForFlag_safe(interrupt_flag, timeout_s, interrupt)
